from collections.abc import Mapping
from datetime import datetime
from typing import Any, Literal
from unittest.mock import MagicMock

import pytest
from cognite.client.data_classes import (
    Asset,
    CountAggregate,
    DataSet,
    UserProfile,
    UserProfileList,
)
from cognite.client.data_classes.aggregations import CountValue
from cognite.client.data_classes.data_modeling import (
    ContainerId,
    MappedProperty,
    NodeList,
    Space,
    SpaceList,
    Text,
    View,
    ViewId,
    ViewList,
)
from cognite.client.data_classes.data_modeling.statistics import SpaceStatistics, SpaceStatisticsList
from cognite.client.data_classes.raw import Database, DatabaseList, Table, TableList
from questionary import Choice

from cognite_toolkit._cdf_tk.client.data_classes.canvas import CANVAS_INSTANCE_SPACE, Canvas
from cognite_toolkit._cdf_tk.client.data_classes.charts import Chart, ChartList
from cognite_toolkit._cdf_tk.client.data_classes.charts_data import ChartData
from cognite_toolkit._cdf_tk.client.data_classes.migration import ResourceViewMapping
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.exceptions import ToolkitMissingResourceError, ToolkitValueError
from cognite_toolkit._cdf_tk.utils.aggregators import AssetCentricAggregator
from cognite_toolkit._cdf_tk.utils.interactive_select import (
    AssetCentricDestinationSelect,
    AssetInteractiveSelect,
    DataModelingSelect,
    EventInteractiveSelect,
    FileMetadataInteractiveSelect,
    InteractiveCanvasSelect,
    InteractiveChartSelect,
    RawTableInteractiveSelect,
    ResourceViewMappingInteractiveSelect,
    TimeSeriesInteractiveSelect,
)
from tests.test_unit.utils import MockQuestionary


class TestInteractiveSelect:
    def test_interactive_select_assets(self, monkeypatch) -> None:
        def select_hierarchy(choices: list[Choice]) -> str:
            assert len(choices) == 2
            return choices[1].value

        def select_data_set(choices: list[Choice]) -> list[str]:
            assert len(choices) == 3 + 1  # +1 for "All Data Sets" option
            return [choices[3].value]

        answers = ["Hierarchy", select_hierarchy, select_data_set]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(AssetInteractiveSelect.__module__, monkeypatch, answers),
        ):
            selector = AssetInteractiveSelect(client, "test_operation")
            client.assets.list.return_value = [Asset(id=1, external_id="Root1"), Asset(id=2, external_id="Root2")]
            aggregator = MagicMock(spec=AssetCentricAggregator)
            aggregator.count.return_value = 1000
            aggregator.used_data_sets.return_value = ["dataset1", "dataset2", "dataset3"]
            selector._aggregator = aggregator
            client.data_sets.retrieve_multiple.return_value = [
                DataSet(id=1, external_id="dataset1"),
                DataSet(id=2, external_id="dataset2"),
                DataSet(id=3, external_id="dataset3"),
            ]

            selected_hierarchy, selected_dataset = selector.select_hierarchies_and_data_sets()

        assert selected_hierarchy == ["Root2"]
        assert selected_dataset == ["dataset3"]

    def test_interactive_select_filemetadata(self, monkeypatch) -> None:
        def select_data_set(choices: list[Choice]) -> str:
            assert len(choices) == 3
            return choices[2].value

        def select_hierarchy(choices: list[Choice]) -> list[str]:
            assert len(choices) == 2 + 1  # +1 for "All Hierarchies" option
            return [choices[2].value]

        answers = ["Data Set", select_data_set, select_hierarchy]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(FileMetadataInteractiveSelect.__module__, monkeypatch, answers),
        ):
            client.data_sets.list.return_value = [
                DataSet(id=1, external_id="dataset1", name="Dataset 1"),
                DataSet(id=2, external_id="dataset2", name="Dataset 2"),
                DataSet(id=3, external_id="dataset3", name="Dataset 3"),
            ]
            client.assets.list.return_value = [
                Asset(id=1, external_id="Root1", name="Root 1"),
                Asset(id=2, external_id="Root2", name="Root 2"),
            ]
            client.files.aggregate.return_value = [CountAggregate(100)]
            selector = FileMetadataInteractiveSelect(client, "test_operation")
            selected_hierarchy, selected_dataset = selector.select_hierarchies_and_data_sets()

        assert selected_hierarchy == ["Root2"]
        assert selected_dataset == ["dataset3"]

    def test_interactive_select_filemetadata_empty_cdf(self, monkeypatch) -> None:
        def select_data_set(choices: list[Choice]) -> None:
            assert len(choices) == 0
            return None

        def select_hierarchy(choices: list[Choice]) -> list[str]:
            assert len(choices) == 0
            return []

        answers = ["Data Set", select_data_set, select_hierarchy]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(FileMetadataInteractiveSelect.__module__, monkeypatch, answers),
        ):
            client.data_sets.list.return_value = []
            client.assets.list.return_value = []
            client.files.aggregate.return_value = [CountAggregate(100)]
            selector = FileMetadataInteractiveSelect(client, "test_operation")
            with pytest.raises(ToolkitValueError) as exc_info:
                _ = selector.select_hierarchies_and_data_sets()

        assert str(exc_info.value) == "No data Set available to select."

    def test_interactive_select_timeseries(self, monkeypatch) -> None:
        def select_data_set(choices: list[Choice]) -> str:
            assert len(choices) == 3
            return choices[2].value

        def select_hierarchy(choices: list[Choice]) -> list[str]:
            assert len(choices) == 2 + 1  # +1 for "All Hierarchies" option
            return [choices[2].value]

        answers = ["Data Set", select_data_set, select_hierarchy]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(TimeSeriesInteractiveSelect.__module__, monkeypatch, answers),
        ):
            client.data_sets.list.return_value = [
                DataSet(id=1, external_id="dataset1", name="Dataset 1"),
                DataSet(id=2, external_id="dataset2", name="Dataset 2"),
                DataSet(id=3, external_id="dataset3", name="Dataset 3"),
            ]
            client.assets.list.return_value = [
                Asset(id=1, external_id="Root1", name="Root 1"),
                Asset(id=2, external_id="Root2", name="Root 2"),
            ]
            client.time_series.aggregate_count.return_value = 100
            selector = TimeSeriesInteractiveSelect(client, "test_operation")
            selected_hierarchy, selected_dataset = selector.select_hierarchies_and_data_sets()

        assert selected_hierarchy == ["Root2"]
        assert selected_dataset == ["dataset3"]

    def test_interactive_select_events(self, monkeypatch) -> None:
        def select_data_set(choices: list[Choice]) -> str:
            assert len(choices) == 3
            return choices[2].value

        def select_hierarchy(choices: list[Choice]) -> list[str]:
            assert len(choices) == 2 + 1  # +1 for "All Hierarchies" option
            return [choices[2].value]

        answers = ["Data Set", select_data_set, select_hierarchy]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(EventInteractiveSelect.__module__, monkeypatch, answers),
        ):
            client.data_sets.list.return_value = [
                DataSet(id=1, external_id="dataset1", name="Dataset 1"),
                DataSet(id=2, external_id="dataset2", name="Dataset 2"),
                DataSet(id=3, external_id="dataset3", name="Dataset 3"),
            ]
            client.assets.list.return_value = [
                Asset(id=1, external_id="Root1", name="Root 1"),
                Asset(id=2, external_id="Root2", name="Root 2"),
            ]
            client.events.aggregate_count.return_value = 100
            selector = EventInteractiveSelect(client, "test_operation")
            selected_hierarchy, selected_dataset = selector.select_hierarchies_and_data_sets()

        assert selected_hierarchy == ["Root2"]
        assert selected_dataset == ["dataset3"]

    def test_asset_centric_destination_select(self, monkeypatch) -> None:
        def select_destination(choices: list[Choice]) -> str:
            assert len(choices) == len(AssetCentricDestinationSelect.valid_destinations)
            return choices[2].value

        answers = [select_destination]
        with MockQuestionary(AssetCentricDestinationSelect.__module__, monkeypatch, answers):
            selected = AssetCentricDestinationSelect.select()

        assert selected == AssetCentricDestinationSelect.valid_destinations[2]

    def test_asset_centric_destination_invalid_destination(self) -> None:
        with pytest.raises(ValueError) as exc_info:
            AssetCentricDestinationSelect.validate("invalid_destination")

        assert "Invalid destination type: 'invalid_destination'." in str(exc_info.value)

    def test_select_data_set(self, monkeypatch):
        def select_data_set(choices) -> str:
            assert len(choices) == 2
            selection = choices[1].value
            assert isinstance(selection, str)
            return selection

        answers = [select_data_set]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(AssetInteractiveSelect.__module__, monkeypatch, answers),
        ):
            selector = AssetInteractiveSelect(client, "test")
            aggregator = MagicMock(spec=AssetCentricAggregator)
            aggregator.count.return_value = 1000
            selector._aggregator = aggregator

            client.data_sets.list.return_value = [
                DataSet(id=1, external_id="ds1", name="DataSet 1"),
                DataSet(id=2, external_id="ds2", name="DataSet 2"),
            ]
            result = selector.select_data_set()
        assert result == "ds2"

    def test_select_data_sets_allow_empty(self, monkeypatch) -> None:
        def select_data_sets(choices) -> str | None:
            assert len(choices) == 2 + 1  # +1 for "All Data Sets" option
            assert choices[0].title.startswith("All")
            return choices[0].value

        answers = [select_data_sets]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(AssetInteractiveSelect.__module__, monkeypatch, answers),
        ):
            selector = AssetInteractiveSelect(client, "test")
            aggregator = MagicMock(spec=AssetCentricAggregator)
            aggregator.count.return_value = 1000
            selector._aggregator = aggregator

            client.data_sets.list.return_value = [
                DataSet(id=1, external_id="ds1", name="DataSet 1"),
                DataSet(id=2, external_id="ds2", name="DataSet 2"),
            ]

            result = selector.select_data_set(allow_empty=True)
        assert result is None

    def test_select_data_sets(self, monkeypatch):
        def select_data_sets(choices) -> str:
            assert len(choices) == 2
            selection = choices[0].value
            assert isinstance(selection, str)
            return selection

        answers = [select_data_sets]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(AssetInteractiveSelect.__module__, monkeypatch, answers),
        ):
            selector = AssetInteractiveSelect(client, "test")
            aggregator = MagicMock(spec=AssetCentricAggregator)
            aggregator.count.return_value = 1000
            selector._aggregator = aggregator

            client.data_sets.list.return_value = [
                DataSet(id=1, external_id="ds1", name="DataSet 1"),
                DataSet(id=2, external_id="ds2", name="DataSet 2"),
            ]

            result = selector.select_data_sets()
        assert result == "ds1"

    def test_select_hierarchy(self, monkeypatch):
        def select_hierarchy(choices) -> str:
            assert len(choices) == 2
            selection = choices[0].value
            assert isinstance(selection, str)
            return selection

        answers = [select_hierarchy]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(AssetInteractiveSelect.__module__, monkeypatch, answers),
        ):
            selector = AssetInteractiveSelect(client, "test")
            aggregator = MagicMock(spec=AssetCentricAggregator)
            aggregator.count.return_value = 1000
            selector._aggregator = aggregator

            client.assets.list.return_value = [
                Asset(id=1, external_id="root1", name="Root 1"),
                Asset(id=2, external_id="root2", name="Root 2"),
            ]

            result = selector.select_hierarchy()
        assert result == "root1"

    def test_select_hierarchy_allow_empty(self, monkeypatch) -> None:
        def select_hierarchy(choices) -> str | None:
            assert len(choices) == 2 + 1
            # +1 for "All Hierarchies" option
            assert choices[0].title.startswith("All")
            return choices[0].value

        answers = [select_hierarchy]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(AssetInteractiveSelect.__module__, monkeypatch, answers),
        ):
            selector = AssetInteractiveSelect(client, "test")
            aggregator = MagicMock(spec=AssetCentricAggregator)
            aggregator.count.return_value = 1000
            selector._aggregator = aggregator

            client.assets.list.return_value = [
                Asset(id=1, external_id="root1", name="Root 1"),
                Asset(id=2, external_id="root2", name="Root 2"),
            ]

            result = selector.select_hierarchy(allow_empty=True)
        assert result is None

    def test_select_hierarchies(self, monkeypatch):
        def select_hierarchies(choices) -> list[str]:
            assert len(choices) == 2
            return [choices[1].value]

        answers = [select_hierarchies]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(AssetInteractiveSelect.__module__, monkeypatch, answers),
        ):
            selector = AssetInteractiveSelect(client, "test")
            aggregator = MagicMock(spec=AssetCentricAggregator)
            aggregator.count.return_value = 1000
            selector._aggregator = aggregator
            client.assets.list.return_value = [
                Asset(id=1, external_id="root1", name="Root 1"),
                Asset(id=2, external_id="root2", name="Root 2"),
            ]
            result = selector.select_hierarchies()
        assert result == ["root2"]


class TestRawTableSelect:
    def test_interactive_select_raw_table(self, monkeypatch) -> None:
        def select_database(choices: list[Choice]) -> str:
            assert len(choices) == 3
            return choices[2].value

        def select_tables(choices: list[Choice]) -> list[RawTable]:
            assert len(choices) == 23
            return [choices[i].value for i in range(0, 24, 2)]

        answers = [select_database, select_tables]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(RawTableInteractiveSelect.__module__, monkeypatch, answers),
        ):
            client.raw.databases.list.return_value = DatabaseList([Database(name=f"Database{i}") for i in range(1, 4)])
            client.raw.tables.list.return_value = TableList([Table(name=f"Table{i}") for i in range(1, 24)])
            selected = RawTableInteractiveSelect(client, "test_operation").select_tables()

        assert selected == [RawTable("Database3", f"Table{i}") for i in range(1, 24, 2)]


class TestInteractiveCanvasSelect:
    @pytest.mark.parametrize(
        "selected_cdf, answers, expected_selected, expected_options",
        [
            pytest.param({}, ["All Canvases"], [], None, id="No canvases in CDF"),
            pytest.param(
                {"Public1", "Public2", "Private1", "Private2"},
                ["All Canvases"],
                ["Public1", "Public2", "Private1", "Private2"],
                None,
                id="All canvases selected",
            ),
            pytest.param(
                {"Public1", "Public2"},
                ["All public Canvases"],
                ["Public1", "Public2"],
                None,
                id="All public canvases selected",
            ),
            pytest.param(
                {"Public1", "Public2"},
                ["Selected public Canvases", {"Public1"}],
                ["Public1"],
                2,
                id="Selected public canvases",
            ),
            pytest.param(
                {"Public1", "Public2", "Private1", "Private2"},
                ["All by given user", "Marge Simpson (marge)"],
                ["Public2", "Private1"],
                None,
                id="All by given user",
            ),
            pytest.param(
                {"Public1", "Public2", "Private1", "Private2"},
                ["Selected by given user", "Marge Simpson (marge)", {"Public2"}],
                ["Public2"],
                2,
                id="Selected by given user",
            ),
        ],
    )
    def test_interactive_selection(
        self,
        selected_cdf: set[str],
        answers: list,
        expected_selected: list[str],
        expected_options: int | None,
        monkeypatch,
    ) -> None:
        default_args = dict(
            space=CANVAS_INSTANCE_SPACE,
            version=1,
            last_updated_time=1,
            created_time=1,
            updated_by="Irrelevant",
            updated_at=datetime.now(),
        )
        cdf_canvases = [
            Canvas(external_id="Public1", name="Canvas 1", visibility="public", created_by="homer", **default_args),
            Canvas(external_id="Public2", name="Canvas 2", visibility="public", created_by="marge", **default_args),
            Canvas(external_id="Private1", name="Private 1", visibility="private", created_by="marge", **default_args),
            Canvas(external_id="Private2", name="Private 2", visibility="private", created_by="homer", **default_args),
        ]
        first_answer_by_choice_title = {c.title: c.value for c in InteractiveCanvasSelect.opening_choices}
        assert len(answers) >= 1, "At least one answer is required to select a canvas"
        assert answers[0] in first_answer_by_choice_title, "Bug in test data. First answer must be a choice title"
        if "user" in answers[0] and len(answers) >= 2:
            user_title = answers[1]

            def select_user(choices: list[Choice]) -> str:
                assert len(choices) == 2
                user_choice = next((c for c in choices if c.title == user_title), None)
                assert user_choice is not None, f"Bug in test data. User choice '{user_title}' not found in choices"
                return user_choice.value

            answers[1] = select_user
        answers[0] = first_answer_by_choice_title[answers[0]]

        if expected_options is not None:
            # Last question is which canvases to select.
            last_selection = answers[-1]

            def select_canvases(choices: list[Choice]) -> list[str]:
                assert len(choices) == expected_options, (
                    f"Expected {expected_options} choices, but got {len(choices)} choices: {choices}"
                )
                return [choice.value for choice in choices if choice.value in last_selection]

            answers[-1] = select_canvases

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(InteractiveCanvasSelect.__module__, monkeypatch, answers),
        ):
            client.canvas.list.return_value = NodeList[Canvas](
                [canvas for canvas in cdf_canvases if canvas.external_id in selected_cdf]
            )
            client.iam.user_profiles.list.return_value = UserProfileList(
                [
                    UserProfile(user_identifier="homer", display_name="Homer Simpson", last_updated_time=1),
                    UserProfile(user_identifier="marge", display_name="Marge Simpson", last_updated_time=1),
                ]
            )
            selector = InteractiveCanvasSelect(client)
            selected_external_ids = selector.select_external_ids()

        assert selected_external_ids == expected_selected

    def test_interactive_abort_selection(self, monkeypatch) -> None:
        answers = [None]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(InteractiveCanvasSelect.__module__, monkeypatch, answers),
        ):
            selector = InteractiveCanvasSelect(client)
            with pytest.raises(ToolkitValueError) as exc_info:
                selector.select_external_ids()
        assert str(exc_info.value) == "No Canvas selection made. Aborting."


class TestInteractiveChartSelect:
    @pytest.mark.parametrize(
        "selected_cdf, answers, expected_selected, expected_options",
        [
            pytest.param([], ["All public Charts"], [], None, id="No charts in CDF"),
            pytest.param(
                ["homer1", "homer2", "marge1", "marge2"],
                ["All public Charts"],
                ["homer1", "homer2", "marge1", "marge2"],
                None,
                id="All public charts selected",
            ),
            pytest.param(
                ["homer1", "homer2", "marge1", "marge2"],
                ["Selected public Charts", {"homer1"}],
                ["homer1"],
                4,
                id="Selected public charts",
            ),
            pytest.param(
                ["homer1", "homer2", "marge1", "marge2"],
                ["All owned by given user", "Marge Simpson (marge)"],
                ["marge1", "marge2"],
                None,
                id="All by given user",
            ),
            pytest.param(
                ["homer1", "homer2", "marge1", "marge2"],
                ["Selected owned by given user", "Marge Simpson (marge)", {"marge2"}],
                ["marge2"],
                2,
                id="Selected by given user",
            ),
        ],
    )
    def test_interactive_selection(
        self,
        selected_cdf: list[str],
        answers: list,
        expected_selected: list[str],
        expected_options: int | None,
        monkeypatch,
    ) -> None:
        # Use real Chart and ChartData objects instead of DummyChart
        default_args = dict(
            created_time=1,
            last_updated_time=1,
            visibility="PUBLIC",
        )
        cdf_charts = [
            Chart(external_id="homer1", data=ChartData(name="Homer 1"), owner_id="homer", **default_args),
            Chart(external_id="homer2", data=ChartData(name="Homer 2"), owner_id="homer", **default_args),
            Chart(external_id="marge1", data=ChartData(name="Marge 1"), owner_id="marge", **default_args),
            Chart(external_id="marge2", data=ChartData(name="Marge 2"), owner_id="marge", **default_args),
        ]
        # Map answer titles to opening_choices values
        first_answer_by_choice_title = {c.title: c.value for c in InteractiveChartSelect.opening_choices}
        assert len(answers) >= 1, "At least one answer is required to select a chart"
        assert answers[0] in first_answer_by_choice_title, "Bug in test data. First answer must be a choice title"

        if "user" in answers[0] and len(answers) >= 2:
            user_title = answers[1]

            def select_user(choices: list[Choice]) -> str:
                assert len(choices) == 2
                user_choice = next((c for c in choices if c.title == user_title), None)
                assert user_choice is not None, f"Bug in test data. User choice '{user_title}' not found in choices"
                return user_choice.value

            answers[1] = select_user
        answers[0] = first_answer_by_choice_title[answers[0]]
        if expected_options is not None:
            last_selection = answers[-1]

            def select_charts(choices: list[Choice]) -> list[str]:
                assert len(choices) == expected_options, (
                    f"Expected {expected_options} choices, but got {len(choices)} choices: {choices}"
                )
                return [choice.value for choice in choices if choice.value in last_selection]

            answers[-1] = select_charts
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(InteractiveChartSelect.__module__, monkeypatch, answers),
        ):
            # Only include charts whose external_id is in selected_cdf
            client.charts.list.return_value = ChartList(
                [chart for chart in cdf_charts if chart.external_id in selected_cdf]
            )
            client.iam.user_profiles.list.return_value = UserProfileList(
                [
                    UserProfile(user_identifier="homer", display_name="Homer Simpson", last_updated_time=1),
                    UserProfile(user_identifier="marge", display_name="Marge Simpson", last_updated_time=1),
                ]
            )
            selector = InteractiveChartSelect(client)
            selected_external_ids = selector.select_external_ids()
        assert selected_external_ids == expected_selected

    def test_interactive_abort_selection(self, monkeypatch) -> None:
        answers = [None]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(InteractiveChartSelect.__module__, monkeypatch, answers),
        ):
            selector = InteractiveChartSelect(client)
            with pytest.raises(ToolkitValueError) as exc_info:
                selector.select_external_ids()
        assert str(exc_info.value) == "No Chart selection made. Aborting."


class TestDataModelingInteractiveSelect:
    DEFAULT_SPACE_ARGS: Mapping = dict(
        description="Test Space",
        name="Test Space",
        created_time=1,
        last_updated_time=1,
        is_global=False,
    )
    DEFAULT_VIEW_ARGS: Mapping = dict(
        properties={},
        last_updated_time=1,
        created_time=1,
        description=None,
        name=None,
        filter=None,
        implements=None,
        writable=True,
        used_for="node",
        is_global=False,
    )

    @pytest.mark.parametrize(
        "multiselect, space, expected",
        [
            pytest.param(False, None, "view2", id="Single view, select space"),
            pytest.param(True, None, {"view1", "view3"}, id="Multiple views, select space"),
            pytest.param(False, "space1", "view3", id="Single view, given space"),
            pytest.param(True, "space1", {"view4"}, id="Multiple views, given space"),
        ],
    )
    def test_select_view(
        self, multiselect: bool, space: str | None, expected: str | set[str], monkeypatch: pytest.MonkeyPatch
    ) -> None:
        assert (multiselect and isinstance(expected, set)) or (not multiselect and isinstance(expected, str)), (
            "Test data error: expected type does not match multiselect mode"
        )
        spaces = [
            Space(space="space1", **self.DEFAULT_SPACE_ARGS),
            Space(space="space2", **self.DEFAULT_SPACE_ARGS),
        ]
        views = [
            View(space="space1", external_id="view1", version="1", **self.DEFAULT_VIEW_ARGS),
            View(space="space1", external_id="view2", version="1", **self.DEFAULT_VIEW_ARGS),
            View(space="space1", external_id="view3", version="1", **self.DEFAULT_VIEW_ARGS),
            View(space="space1", external_id="view4", version="1", **self.DEFAULT_VIEW_ARGS),
        ]
        space_stats = SpaceStatisticsList([SpaceStatistics(space.space, 0, 1, 0, 0, 0, 0, 0) for space in spaces])
        answers: list[Any] = []
        if space is None:
            answers.append(spaces[0])
        if multiselect:
            answers.append([view for view in views if view.external_id in expected])
        else:
            answers.append(next(view for view in views if view.external_id == expected))

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, answers),
        ):
            if space is None:
                client.data_modeling.spaces.list.return_value = SpaceList(spaces)
            client.data_modeling.views.list.return_value = ViewList(views)
            client.data_modeling.statistics.spaces.list.return_value = space_stats

            selector = DataModelingSelect(client, "test_operation")
            selected_view = selector.select_view(multiselect=multiselect, space=space)
        if multiselect:
            assert isinstance(selected_view, ViewList)
            assert {view.external_id for view in selected_view} == expected
        else:
            assert isinstance(selected_view, View)
            assert selected_view.external_id == expected

    def test_select_view_mapped_container(self, monkeypatch: pytest.MonkeyPatch) -> None:
        space = Space(space="space1", **self.DEFAULT_SPACE_ARGS)
        default_view_args = dict(self.DEFAULT_VIEW_ARGS)
        default_view_args["properties"] = {
            "name": MappedProperty(
                container=ContainerId(space="space1", external_id="container1"),
                container_property_identifier="name",
                type=Text(),
                nullable=True,
                immutable=False,
                auto_increment=False,
            )
        }
        views = [
            View(space="space1", external_id="view1", version="1", **self.DEFAULT_VIEW_ARGS),
            View(space="space1", external_id="view2", version="1", **default_view_args),
        ]
        space_stats = SpaceStatisticsList([SpaceStatistics(space.space, 0, 1, 0, 0, 0, 0, 0)])

        def select_view(choices: list[Choice]) -> View:
            assert len(choices) == 1, "Expected one view to be filtered out."
            return choices[0].value

        answers = [space, select_view]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, answers),
        ):
            client.data_modeling.spaces.list.return_value = SpaceList([space])
            client.data_modeling.statistics.spaces.list.return_value = space_stats
            client.data_modeling.views.list.return_value = ViewList(views)
            selector = DataModelingSelect(client, "test_operation")
            selected_view = selector.select_view(mapped_container=ContainerId("space1", "container1"))

        assert selected_view.external_id == "view2"

    def test_select_no_schema_space_found(self, monkeypatch) -> None:
        space = Space(space="space1", **self.DEFAULT_SPACE_ARGS)
        space_stats = [SpaceStatistics(space.space, 0, 0, 0, 0, 0, 0, 0)]
        answers = [space]  # Direct string answer
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, answers),
        ):
            client.data_modeling.spaces.list.return_value = SpaceList([space])
            client.data_modeling.statistics.spaces.list.return_value = SpaceStatisticsList(space_stats)
            client.data_modeling.views.list.return_value = []
            selector = DataModelingSelect(client, "test_operation")
            with pytest.raises(ToolkitMissingResourceError) as exc_info:
                selector.select_view()
            assert str(exc_info.value) == "No spaces with schema (containers, views, or data models) found."

    @pytest.mark.parametrize(
        "answer, input",
        [
            pytest.param("node", "all", id="View Used for all selecting node"),
            pytest.param("edge", None, id="Selecting instance type without view"),
        ],
    )
    def test_select_instance_type(
        self, answer: str, input: Literal["node", "edge", "all"] | None, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        answers = [answer]  # Direct string answer
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, answers),
        ):
            selector = DataModelingSelect(client, "test_operation")
            instance_type = selector.select_instance_type(input)

        assert instance_type == answer

    def test_select_single_schema_space(self, monkeypatch) -> None:
        spaces = [
            Space(space="space1", **self.DEFAULT_SPACE_ARGS),
            Space(space="space2", **self.DEFAULT_SPACE_ARGS),
        ]
        stats = [SpaceStatistics(space.space, 0, 1, 0, 0, 0, 0, 0) for space in spaces]
        answers = [spaces[1]]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, answers),
        ):
            client.data_modeling.spaces.list.return_value = SpaceList(spaces)
            client.data_modeling.statistics.spaces.list.return_value = SpaceStatisticsList(stats)
            selector = DataModelingSelect(client, "test_operation")
            selected_space = selector.select_schema_space(include_global=True)

        assert selected_space.space == "space2"

    def test_select_instance_spaces_one_space_with_instances(self, monkeypatch) -> None:
        def mock_aggregate(view_id, count, instance_type, space):
            if space == "space1":
                return CountValue("externalId", 5)
            return CountValue("externalId", 0)

        with monkeypatch_toolkit_client() as client:
            client.data_modeling.spaces.list.return_value = SpaceList(
                [
                    Space(space="space1", **self.DEFAULT_SPACE_ARGS),
                    Space(space="space2", **self.DEFAULT_SPACE_ARGS),
                ]
            )
            client.data_modeling.instances.aggregate.side_effect = mock_aggregate
            client.data_modeling.statistics.project().concurrent_read_limit = 2

            selector = DataModelingSelect(client, "test_operation")
            selected_spaces = selector.select_instance_space(True, ViewId("space1", "view1", "1"), "node")

        assert selected_spaces == ["space1"]

    def test_select_instance_spaces_multiple_spaces(self, monkeypatch) -> None:
        spaces = [
            Space(space="space1", **self.DEFAULT_SPACE_ARGS),
            Space(space="space2", **self.DEFAULT_SPACE_ARGS),
            Space(space="space3", **self.DEFAULT_SPACE_ARGS),
        ]

        def select_space(choices: list[Choice]) -> list[str]:
            assert len(choices) == 3
            return [choices[0].value, choices[2].value]

        answers = [select_space]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, answers),
        ):
            client.data_modeling.spaces.list.return_value = SpaceList(spaces)
            client.data_modeling.instances.aggregate.return_value = CountValue("externalId", 5)
            client.data_modeling.statistics.project().concurrent_read_limit = 6

            selector = DataModelingSelect(client, "test_operation")
            selected_spaces = selector.select_instance_space(True, ViewId("space1", "view1", "1"), "node")

        assert selected_spaces == ["space1", "space3"]

    def test_select_instance_spaces_no_instances(self, monkeypatch) -> None:
        with monkeypatch_toolkit_client() as client:
            client.data_modeling.spaces.list.return_value = SpaceList(
                [
                    Space(space="space1", **self.DEFAULT_SPACE_ARGS),
                    Space(space="space2", **self.DEFAULT_SPACE_ARGS),
                ]
            )
            client.data_modeling.instances.aggregate.return_value = CountValue("externalId", 0)
            client.data_modeling.statistics.project().concurrent_read_limit = 2

            selector = DataModelingSelect(client, "test_operation")
            with pytest.raises(ToolkitMissingResourceError) as exc_info:
                selector.select_instance_space(True, ViewId("space1", "view1", "1"), "node")

            assert str(exc_info.value) == (
                "No instances found in any space for the view "
                "ViewId(space='space1', external_id='view1', version='1') with instance type 'node'."
            )

    def test_select_space_type(self, monkeypatch) -> None:
        answers = ["schema"]  # Direct string answer
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, answers),
        ):
            selector = DataModelingSelect(client, "test_operation")
            space_type = selector.select_space_type()

        assert space_type == "schema"

    def test_select_empty_spaces(self, monkeypatch) -> None:
        spaces = [
            Space(space="space1", **self.DEFAULT_SPACE_ARGS),
            Space(space="space2", **self.DEFAULT_SPACE_ARGS),
            Space(space="space3", **self.DEFAULT_SPACE_ARGS),
        ]

        # Set up space statistics to make spaces 1 and 3 empty
        space_stats = [
            SpaceStatistics("space1", 0, 0, 0, 0, 0, 0, 0),  # Empty space
            SpaceStatistics("space2", 5, 2, 1, 0, 0, 0, 0),  # Non-empty space
            SpaceStatistics("space3", 0, 0, 0, 0, 0, 0, 0),  # Empty space
        ]

        def select_spaces(choices: list[Choice]) -> list[str]:
            assert len(choices) == 2  # Only the empty spaces should be presented
            return [choices[1].value]  # Select space3

        answers = [select_spaces]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, answers),
        ):
            client.data_modeling.spaces.list.return_value = SpaceList(spaces)
            client.data_modeling.statistics.spaces.list.return_value = SpaceStatisticsList(space_stats)

            selector = DataModelingSelect(client, "test_operation")
            selected_spaces = selector.select_empty_spaces()

        assert selected_spaces == ["space3"]

    def test_select_empty_spaces_single_space(self, monkeypatch) -> None:
        spaces = [
            Space(space="space1", **self.DEFAULT_SPACE_ARGS),
            Space(space="space2", **self.DEFAULT_SPACE_ARGS),
        ]

        # Only space1 is empty
        space_stats = [
            SpaceStatistics("space1", 0, 0, 0, 0, 0, 0, 0),  # Empty space
            SpaceStatistics("space2", 5, 2, 1, 0, 0, 0, 0),  # Non-empty space
        ]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, []),
        ):
            client.data_modeling.spaces.list.return_value = SpaceList(spaces)
            client.data_modeling.statistics.spaces.list.return_value = SpaceStatisticsList(space_stats)

            selector = DataModelingSelect(client, "test_operation")
            selected_spaces = selector.select_empty_spaces()

        assert selected_spaces == ["space1"]

    def test_select_empty_spaces_no_spaces(self, monkeypatch) -> None:
        spaces = [
            Space(space="space1", **self.DEFAULT_SPACE_ARGS),
            Space(space="space2", **self.DEFAULT_SPACE_ARGS),
        ]

        # No empty spaces
        space_stats = [
            SpaceStatistics("space1", 5, 0, 0, 0, 0, 0, 0),
            SpaceStatistics("space2", 0, 2, 1, 0, 0, 0, 0),
        ]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, []),
        ):
            client.data_modeling.spaces.list.return_value = SpaceList(spaces)
            client.data_modeling.statistics.spaces.list.return_value = SpaceStatisticsList(space_stats)

            selector = DataModelingSelect(client, "test_operation")
            with pytest.raises(ToolkitMissingResourceError) as exc_info:
                selector.select_empty_spaces()

            assert str(exc_info.value) == "No empty spaces found."

    def test_select_instance_spaces(self, monkeypatch) -> None:
        spaces = [
            Space(space="space1", **self.DEFAULT_SPACE_ARGS),
            Space(space="space2", **self.DEFAULT_SPACE_ARGS),
            Space(space="space3", **self.DEFAULT_SPACE_ARGS),
        ]

        # Set up space statistics with different instance counts
        space_stats = [
            SpaceStatistics("space1", 0, 0, 0, 0, 0, 10, 0),  # 10 nodes
            SpaceStatistics("space2", 0, 0, 0, 0, 0, 0, 0),  # No instances
            SpaceStatistics("space3", 0, 0, 0, 5, 0, 5, 0),  # 5 nodes, 5 edges
        ]

        def select_spaces(choices: list[Choice]) -> list[str]:
            assert len(choices) == 2  # Only spaces with instances should be presented
            return [choices[1].value]  # Select space3

        answers = [select_spaces]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, answers),
        ):
            client.data_modeling.spaces.list.return_value = SpaceList(spaces)
            client.data_modeling.statistics.spaces.list.return_value = SpaceStatisticsList(space_stats)

            selector = DataModelingSelect(client, "test_operation")
            selected_spaces = selector.select_instance_space()

        assert selected_spaces == ["space3"]

    def test_select_instance_spaces_without_view_or_instance_type_single_space(self, monkeypatch) -> None:
        spaces = [
            Space(space="space1", **self.DEFAULT_SPACE_ARGS),
            Space(space="space2", **self.DEFAULT_SPACE_ARGS),
        ]

        # Only space1 has instances
        space_stats = [
            SpaceStatistics("space1", 0, 0, 0, 5, 0, 2, 0),  # Has instances
            SpaceStatistics("space2", 0, 0, 0, 0, 0, 0, 0),  # No instances
        ]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, []),
        ):
            client.data_modeling.spaces.list.return_value = SpaceList(spaces)
            client.data_modeling.statistics.spaces.list.return_value = SpaceStatisticsList(space_stats)

            selector = DataModelingSelect(client, "test_operation")
            selected_spaces = selector.select_instance_space()

        assert selected_spaces == ["space1"]

    def test_select_instance_spaces_without_view_or_instance_type_no_instances(self, monkeypatch) -> None:
        spaces = [
            Space(space="space1", **self.DEFAULT_SPACE_ARGS),
            Space(space="space2", **self.DEFAULT_SPACE_ARGS),
        ]

        # No spaces have instances
        space_stats = [
            SpaceStatistics("space1", 0, 0, 1, 0, 0, 0, 0),  # Only has containers
            SpaceStatistics("space2", 0, 2, 0, 0, 0, 0, 0),  # Only has views
        ]

        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(DataModelingSelect.__module__, monkeypatch, []),
        ):
            client.data_modeling.spaces.list.return_value = SpaceList(spaces)
            client.data_modeling.statistics.spaces.list.return_value = SpaceStatisticsList(space_stats)

            selector = DataModelingSelect(client, "test_operation")
            with pytest.raises(ToolkitMissingResourceError) as exc_info:
                selector.select_instance_space()

            assert "No instances found in any space" in str(exc_info.value)


class TestResourceViewMappingInteractiveSelect:
    def test_interactive_select_resource_view_mapping(self, monkeypatch) -> None:
        def select_resource_view_mapping(choices: list[Choice]) -> ResourceViewMapping:
            assert len(choices) == 2
            selection = choices[1].value
            return selection

        answers = [select_resource_view_mapping]
        with (
            monkeypatch_toolkit_client() as client,
            MockQuestionary(ResourceViewMappingInteractiveSelect.__module__, monkeypatch, answers),
        ):
            selector = ResourceViewMappingInteractiveSelect(client, "test_operation")
            client.migration.resource_view_mapping.list.return_value = NodeList[ResourceViewMapping](
                [
                    ResourceViewMapping(
                        external_id="mapping1",
                        resource_type="asset",
                        view_id=ViewId("space1", "view1", "1"),
                        property_mapping={},
                        last_updated_time=1,
                        created_time=0,
                        version=1,
                    ),
                    ResourceViewMapping(
                        external_id="mapping2",
                        resource_type="asset",
                        view_id=ViewId("space2", "view2", "1"),
                        property_mapping={},
                        last_updated_time=1,
                        created_time=0,
                        version=1,
                    ),
                ]
            )

            result = selector.select_resource_view_mapping("asset")
        assert result.external_id == "mapping2"



import inspect
import re
from pathlib import Path
from typing import Any, TypeVar

from cognite.client.data_classes._base import CogniteObject
from cognite.client.utils._text import to_snake_case
from pydantic import BaseModel, TypeAdapter, ValidationError
from pydantic_core import ErrorDetails

from cognite_toolkit._cdf_tk.data_classes import BuildVariables
from cognite_toolkit._cdf_tk.resource_classes import BaseModelResource
from cognite_toolkit._cdf_tk.tk_warnings import (
    DataSetMissingWarning,
    TemplateVariableWarning,
    WarningList,
)
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning

__all__ = [
    "humanize_validation_error",
    "validate_data_set_is_set",
    "validate_modules_variables",
]


T_BaseModel = TypeVar("T_BaseModel", bound=BaseModel)


def validate_modules_variables(variables: BuildVariables, filepath: Path) -> WarningList:
    """Checks whether the config file has any issues.

    Currently, this checks for:
        * Non-replaced template variables, such as <change_me>.

    Args:
        variables: The variables to check.
        filepath: The filepath of the config.yaml.
    """
    warning_list: WarningList = WarningList()
    pattern = re.compile(r"<.*?>")
    for variable in variables:
        if isinstance(variable.value, str) and pattern.match(variable.value):
            warning_list.append(
                TemplateVariableWarning(filepath, variable.value, variable.key, ".".join(variable.location.parts))
            )
    return warning_list


def validate_data_set_is_set(
    raw: dict[str, Any] | list[dict[str, Any]],
    resource_cls: type[CogniteObject],
    filepath: Path,
    identifier_key: str = "externalId",
) -> WarningList:
    warning_list: WarningList = WarningList()
    signature = inspect.signature(resource_cls.__init__)
    if "data_set_id" not in set(signature.parameters.keys()):
        return warning_list

    if isinstance(raw, list):
        for item in raw:
            warning_list.extend(validate_data_set_is_set(item, resource_cls, filepath, identifier_key))
        return warning_list

    if "dataSetExternalId" in raw or "dataSetId" in raw:
        return warning_list

    value = raw.get(identifier_key, raw.get(to_snake_case(identifier_key), f"No identifier {identifier_key}"))
    warning_list.append(DataSetMissingWarning(filepath, value, identifier_key, resource_cls.__name__))
    return warning_list


def validate_resource_yaml_pydantic(
    data: dict[str, object] | list[dict[str, object]], validation_cls: type[BaseModelResource], source_file: Path
) -> WarningList:
    """Validates the resource given as a dictionary or list of dictionaries with the given pydantic model.

    Args:
        data: The data to validate.
        validation_cls: The pydantic model to validate against.
        source_file: The source file of the resource.

    Returns:
        A list of warnings.

    """
    warning_list: WarningList = WarningList()
    try:
        if isinstance(data, dict):
            validation_cls.model_validate(data, strict=True)
        elif isinstance(data, list):
            TypeAdapter(list[validation_cls]).validate_python(data)  # type: ignore[valid-type]
        else:
            raise ValueError(f"Expected a dictionary or list of dictionaries, got {type(data)}.")
    except ValidationError as e:
        printable_errors = tuple(humanize_validation_error(e))
        if printable_errors:
            warning_list.append(ResourceFormatWarning(source_file, printable_errors))
    return warning_list


def instantiate_class(
    data: dict[str, Any], validation_cls: type[T_BaseModel], source_file: Path, strict: bool = False
) -> T_BaseModel | ResourceFormatWarning:
    """Instantiates a class from a dictionary using the given pydantic model.

    Args:
        data: The data to instantiate the class from.
        validation_cls: The pydantic model to use for instantiation.
        source_file: The source file of the resource.
        strict: Whether to enforce types strictly.

    Returns:
        The instantiated class or a ResourceFormatWarning if validation failed.
    """
    try:
        return validation_cls.model_validate(data, strict=strict)
    except ValidationError as e:
        return ResourceFormatWarning(source_file, tuple(humanize_validation_error(e)))


def humanize_validation_error(error: ValidationError) -> list[str]:
    """Converts a ValidationError to a human-readable format.

    This overwrites the default error messages from Pydantic to be better suited for Toolkit users.

    Args:
        error: The ValidationError to convert.

    Returns:
        A list of human-readable error messages.
    """
    errors: list[str] = []
    item: ErrorDetails

    for item in error.errors(include_input=True, include_url=False):
        loc = item["loc"]
        error_type = item["type"]
        is_metadata_string_value_error = error_type == "string_type" and len(loc) >= 2 and loc[-2] == "metadata"
        if error_type == "missing":
            msg = f"Missing required field: {loc[-1]!r}"
        elif error_type == "extra_forbidden":
            msg = f"Unused field: {loc[-1]!r}"
        elif error_type == "value_error":
            msg = str(item["ctx"]["error"])
        elif error_type in {"literal_error", "list_type"}:
            msg = f"{item['msg']}. Got {item['input']!r}."
        elif is_metadata_string_value_error:
            # We skip metadata string errors. There are multiple reasons for this
            # 1. We often allow non-string metadata values, and parse them to string later. For example, in
            #     ExtractionPipelines.
            # 2. The user often set metadata values to int/bool/float by mistake, but the server accepts these values and
            #     converts them to string. Thus, these are not really errors.
            # 3. The metadata errors flood the output and obscure more important errors, and we see example of
            #     users ignoring all errors because of this (error fatigue).
            continue
        elif error_type == "string_type":
            msg = f"{item['msg']}. Got {item['input']!r} of type {type(item['input']).__name__}. Hint: Use double quotes to force string."
        elif error_type == "model_type":
            model_name = item["ctx"].get("class_name", "unknown")
            msg = f"Input must be an object of type {model_name}. Got {item['input']!r} of type {type(item['input']).__name__}."
        elif error_type in {
            "int_type",
            "bool_type",
            "datetime_type",
            "decimal_type",
            "float_type",
            "time_type",
            "timedelta_type",
            "dict_type",
        }:
            msg = f"{item['msg']}. Got {item['input']!r} of type {type(item['input']).__name__}."
        else:
            # Default to the Pydantic error message
            msg = item["msg"]

        if error_type.endswith("dict_type") and len(loc) > 1:
            # If this is a dict_type error for a JSON field, the location will be:
            #  dict[str,json-or-python[json=any,python=tagged-union[list[...],dict[str,...],str,bool,int,float,none]]]
            #  This is hard to read, so we simplify it to just the field name.
            loc = tuple(["dict" if isinstance(x, str) and "json-or-python" in x else x for x in loc])

        if len(loc) > 1 and error_type in {"extra_forbidden", "missing"}:
            # We skip the last element as this is in the message already
            msg = f"In {as_json_path(loc[:-1])} {msg[:1].casefold()}{msg[1:]}"
        elif len(loc) > 1:
            msg = f"In {as_json_path(loc)} {msg[:1].casefold()}{msg[1:]}"
        elif len(loc) == 1 and isinstance(loc[0], str) and error_type not in {"extra_forbidden", "missing"}:
            msg = f"In field {loc[0]} {msg[:1].casefold()}{msg[1:]}"
        errors.append(msg)
    return errors


def as_json_path(loc: tuple[str | int, ...]) -> str:
    """Converts a location tuple to a JSON path.

    Args:
        loc: The location tuple to convert.

    Returns:
        A JSON path string.
    """
    if not loc:
        return ""
    # +1 to convert from 0-based to 1-based indexing
    prefix = ""
    if isinstance(loc[0], int):
        prefix = "item "

    suffix = ".".join([str(x) if isinstance(x, str) else f"[{x + 1}]" for x in loc]).replace(".[", "[")
    return f"{prefix}{suffix}"



from collections.abc import Sequence

import pytest
import responses
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.extended_timeseries import ExtendedTimeSeriesList


class TestExtendedTimeSeriesAPI:
    @pytest.mark.parametrize(
        "id, external_id, expected_list",
        [
            pytest.param(None, "my_timeseries", False, id="External ID only"),
            pytest.param(123, None, False, id="ID only"),
            pytest.param([123, 456], None, True, id="List of IDs"),
            pytest.param(None, ["my_timeseries", "my_other_timeseries"], True, id="List of External IDs"),
            pytest.param([123, 456], ["my_timeseries", "my_other_timeseries"], True, id="List of IDs and External IDs"),
        ],
    )
    def test_unlink_instance_ids_valid(
        self,
        id: int | Sequence[int] | None,
        external_id: str | SequenceNotStr[str] | None,
        expected_list: bool,
        toolkit_config: ToolkitClientConfig,
    ) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        url = f"{toolkit_config.base_url}/api/v1/projects/test-project/timeseries/unlink-instance-ids"
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.POST,
                url,
                status=200,
                json={"items": [{"externalId": "does-not-matter", "id": 123, "createdTime": 0, "lastUpdatedTime": 0}]},
            )
            result = client.time_series.unlink_instance_ids(id=id, external_id=external_id)

        is_list = isinstance(result, ExtendedTimeSeriesList)
        assert is_list == expected_list, f"Expected result to be a list: {expected_list}, got {is_list}"

    def test_unlink_instance_ids_none_return_none(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        result = client.time_series.unlink_instance_ids(id=None, external_id=None)
        assert result is None

    def test_unlink_instance_ids_invalid(self, toolkit_config: ToolkitClientConfig) -> None:
        client = ToolkitClient(config=toolkit_config, enable_set_pending_ids=True)
        with pytest.raises(
            ValueError, match=r"Cannot specify both id and external_id as single values. Use one or the other."
        ):
            client.time_series.unlink_instance_ids(id=123, external_id="123")



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.datapoint_subscription import DatapointSubscriptionYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_datapoint_subscription_test_cases() -> Iterable:
    yield pytest.param(
        {"name": "Subscription 1"}, {"Missing required field: 'externalId'"}, id="Missing required field: externalId"
    )
    # Test case: more than 10,000 time_series_ids
    yield pytest.param(
        {"externalId": "sub-overlimit", "timeSeriesIds": [f"ts_{i}" for i in range(10_001)]},
        {"The total number of time_series_ids and instance_ids cannot exceed 10000."},
        id="Too many time_series_ids",
    )
    # Test case: both filter and time_series_ids set
    yield pytest.param(
        {"externalId": "sub-filter-and-tsids", "timeSeriesIds": ["ts_1", "ts_2"], "filter": {"some": "filter"}},
        {"Cannot set both filter and time_series_ids/instance_ids."},
        id="Filter and time_series_ids set",
    )


class TestDatapointSubscription:
    @pytest.mark.parametrize("data", list(find_resources("DatapointSubscription")))
    def test_load_valid_dataset(self, data: dict[str, object]) -> None:
        loaded = DatapointSubscriptionYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_datapoint_subscription_test_cases()))
    def test_invalid_datapoint_subscription_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, DatapointSubscriptionYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from .base import BaseModelResource, ToolkitResource


class ThreeDModelIdentifier(BaseModelResource):
    revision_id: int | None = None
    model_id: int | None = None
    name: str | None = None


class ThreeDConfiguration(BaseModelResource):
    full_weight_models: list[ThreeDModelIdentifier] | None = None
    light_weight_models: list[ThreeDModelIdentifier] | None = None


class ResourceFilters(BaseModelResource):
    data_set_external_ids: list[str] | None = None
    asset_subtree_external_ids: list[str] | None = None
    root_asset_external_ids: list[str] | None = None
    external_id_prefix: str | None = None
    spaces: list[str] | None = None


class RootLocationDataFilters(BaseModelResource):
    general: ResourceFilters | None = None
    assets: ResourceFilters | None = None
    files: ResourceFilters | None = None
    timeseries: ResourceFilters | None = None


class ObservationFeatureToggles(BaseModelResource):
    is_enabled: bool | None = None
    is_write_back_enabled: bool | None = None
    notifications_endpoint_external_id: str | None = None
    attachments_endpoint_external_id: str | None = None


class RootLocationFeatureToggles(BaseModelResource):
    three_d: bool | None = None
    trends: bool | None = None
    documents: bool | None = None
    workorders: bool | None = None
    notifications: bool | None = None
    media: bool | None = None
    template_checklist_flow: bool | None = None
    workorder_checklist_flow: bool | None = None
    observations: ObservationFeatureToggles | None = None


class ObservationConfigFieldProperty(BaseModelResource):
    display_title: str | None = None
    display_description: str | None = None
    is_required: bool | None = None


class ObservationConfigDropdownPropertyOption(BaseModelResource):
    id: str | None = None
    value: str | None = None
    label: str | None = None


class ObservationConfigDropdownProperty(ObservationConfigFieldProperty):
    options: list[ObservationConfigDropdownPropertyOption] | None = None


class ObservationsConfig(BaseModelResource):
    files: ObservationConfigFieldProperty | None = None
    description: ObservationConfigFieldProperty | None = None
    asset: ObservationConfigFieldProperty | None = None
    troubleshooting: ObservationConfigFieldProperty | None = None
    type: ObservationConfigDropdownProperty | None = None
    priority: ObservationConfigDropdownProperty | None = None


class RootLocationConfiguration(BaseModelResource):
    asset_external_id: str | None = None
    external_id: str | None = None
    display_name: str | None = None
    three_d_configuration: ThreeDConfiguration | None = None
    data_set_external_id: str | None = None
    template_admins: list[str] | None = None  # list of Group Names
    checklist_admins: list[str] | None = None  # list of Group Names
    app_data_instance_space: str | None = None
    source_data_instance_space: str | None = None
    data_filters: RootLocationDataFilters | None = None
    feature_toggles: RootLocationFeatureToggles | None = None
    observations: ObservationsConfig | None = None


class FeatureConfiguration(BaseModelResource):
    root_location_configurations: list[RootLocationConfiguration] | None = None


class InfieldV1YAML(ToolkitResource):
    external_id: str
    name: str | None = None
    app_data_space_id: str | None = None
    app_data_space_version: str | None = None
    customer_data_space_id: str | None = None
    customer_data_space_version: str | None = None
    feature_configuration: FeatureConfiguration | None = None



import os
import shutil
from dataclasses import dataclass
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from cognite.client import CogniteClient, global_config
from cognite.client.credentials import OAuthClientCredentials
from cognite.client.data_classes import (
    Asset,
    AssetList,
    AssetWrite,
    AssetWriteList,
    DataSet,
    DataSetList,
    DataSetWrite,
    DataSetWriteList,
    Event,
    EventWrite,
    FileMetadata,
    FileMetadataWrite,
    Function,
    FunctionSchedule,
    RowWrite,
    RowWriteList,
    TimeSeries,
    TimeSeriesWrite,
    Transformation,
    TransformationDestination,
    TransformationJob,
    TransformationWrite,
)
from cognite.client.data_classes.data_modeling import Space, SpaceApply
from dotenv import load_dotenv

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase, RawDatabaseList, RawTable, RawTableList
from cognite_toolkit._cdf_tk.commands import CollectCommand
from cognite_toolkit._cdf_tk.cruds import RawDatabaseCRUD, RawTableCRUD
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.cdf import ThrottlerState, raw_row_count
from tests.constants import REPO_ROOT
from tests.test_integration.constants import (
    ASSET_COUNT,
    ASSET_DATASET,
    ASSET_TABLE,
    ASSET_TRANSFORMATION,
    EVENT_COUNT,
    EVENT_DATASET,
    EVENT_TABLE,
    EVENT_TRANSFORMATION,
    FILE_COUNT,
    FILE_DATASET,
    FILE_TABLE,
    FILE_TRANSFORMATION,
    SEQUENCE_COUNT,
    SEQUENCE_DATASET,
    SEQUENCE_TABLE,
    SEQUENCE_TRANSFORMATION,
    TIMESERIES_COUNT,
    TIMESERIES_DATASET,
    TIMESERIES_TABLE,
    TIMESERIES_TRANSFORMATION,
)

THIS_FOLDER = Path(__file__).resolve().parent
TMP_FOLDER = THIS_FOLDER / "tmp"


@pytest.fixture(scope="session")
def toolkit_client_config() -> ToolkitClientConfig:
    load_dotenv(REPO_ROOT / ".env", override=True)
    # Ensure that we do not collect data during tests
    cmd = CollectCommand()
    cmd.execute(action="opt-out")

    cdf_cluster = os.environ["CDF_CLUSTER"]
    credentials = OAuthClientCredentials(
        token_url=os.environ["IDP_TOKEN_URL"],
        client_id=os.environ["IDP_CLIENT_ID"],
        client_secret=os.environ["IDP_CLIENT_SECRET"],
        scopes=[f"https://{cdf_cluster}.cognitedata.com/.default"],
        audience=f"https://{cdf_cluster}.cognitedata.com",
    )
    global_config.disable_pypi_version_check = True
    return ToolkitClientConfig(
        client_name="cdf-toolkit-integration-tests",
        base_url=f"https://{cdf_cluster}.cognitedata.com",
        project=os.environ["CDF_PROJECT"],
        credentials=credentials,
        # We cannot commit auth to WorkflowTrigger and FunctionSchedules.
        is_strict_validation=False,
    )


@pytest.fixture(scope="session")
def cognite_client(toolkit_client_config: ToolkitClientConfig) -> CogniteClient:
    return CogniteClient(toolkit_client_config)


@pytest.fixture(scope="session")
def toolkit_client(toolkit_client_config: ToolkitClientConfig) -> ToolkitClient:
    return ToolkitClient(toolkit_client_config)


@pytest.fixture()
def max_two_workers():
    old = global_config.max_workers
    global_config.max_workers = 2
    yield
    global_config.max_workers = old


@pytest.fixture(scope="session")
def toolkit_client_with_pending_ids(toolkit_client_config: ToolkitClientConfig) -> ToolkitClient:
    """Returns a ToolkitClient configured to enable pending IDs."""
    return ToolkitClient(toolkit_client_config, enable_set_pending_ids=True)


@pytest.fixture(scope="session")
def env_vars(toolkit_client: ToolkitClient) -> EnvironmentVariables:
    env_vars = EnvironmentVariables.create_from_environment()
    # Ensure we use the client above that has CLIENT NAME set to the test name
    env_vars._client = toolkit_client
    return env_vars


@pytest.fixture(scope="session")
def toolkit_space(cognite_client: CogniteClient) -> Space:
    return cognite_client.data_modeling.spaces.apply(SpaceApply(space="toolkit_test_space"))


@pytest.fixture(scope="session")
def toolkit_dataset(cognite_client: CogniteClient) -> DataSet:
    """Returns the dataset name used for toolkit tests."""
    dataset = DataSetWrite(
        external_id="toolkit_tests_dataset", name="Toolkit Test DataSet", description="Toolkit DataSet used in tests"
    )
    retrieved = cognite_client.data_sets.retrieve(external_id=dataset.external_id)
    if retrieved is None:
        return cognite_client.data_sets.create(dataset)
    return retrieved


@pytest.fixture
def build_dir() -> Path:
    pidid = os.getpid()
    build_path = TMP_FOLDER / f"build-{pidid}"
    build_path.mkdir(exist_ok=True, parents=True)
    yield build_path
    shutil.rmtree(build_path, ignore_errors=True)


@pytest.fixture(scope="session")
def dev_cluster_client() -> ToolkitClient | None:
    """Returns a ToolkitClient configured for the development cluster."""
    dev_cluster_env = REPO_ROOT / "dev-cluster.env"
    if not dev_cluster_env.exists():
        pytest.skip("dev-cluster.env file not found, skipping tests that require dev cluster client.")
        return None
    env_content = dev_cluster_env.read_text(encoding="utf-8")
    env_vars = dict(
        line.strip().split("=")
        for line in env_content.splitlines()
        if line.strip() and not line.startswith("#") and "=" in line
    )
    cdf_cluster = env_vars["CDF_CLUSTER"]
    credentials = OAuthClientCredentials(
        token_url=env_vars["IDP_TOKEN_URL"],
        client_id=env_vars["IDP_CLIENT_ID"],
        client_secret=env_vars["IDP_CLIENT_SECRET"],
        scopes=[f"https://{cdf_cluster}.cognitedata.com/.default"],
        audience=f"https://{cdf_cluster}.cognitedata.com",
    )
    config = ToolkitClientConfig(
        client_name="cdf-toolkit-integration-tests",
        base_url=f"https://{cdf_cluster}.cognitedata.com",
        project=env_vars["CDF_PROJECT"],
        credentials=credentials,
        is_strict_validation=False,
    )
    return ToolkitClient(config, enable_set_pending_ids=True)


@pytest.fixture(scope="session")
def dummy_function(cognite_client: CogniteClient) -> Function:
    external_id = "integration_test_function_dummy"

    if existing := cognite_client.functions.retrieve(external_id=external_id):
        return existing

    def handle(client: CogniteClient, data: dict, function_call_info: dict) -> str:
        """
        [requirements]
        cognite-sdk>=7.37.0
        [/requirements]
        """
        print("Print statements will be shown in the logs.")
        print("Running with the following configuration:\n")
        return {
            "data": data,
            "functionInfo": function_call_info,
        }

    return cognite_client.functions.create(
        name="integration_test_function_dummy",
        function_handle=handle,
        external_id="integration_test_function_dummy",
    )


@pytest.fixture
def dummy_schedule(cognite_client: CogniteClient, dummy_function: Function) -> FunctionSchedule:
    name = "integration_test_schedule_dummy"
    if existing_list := cognite_client.functions.schedules.list(name=name):
        if len(existing_list) > 1:
            for existing in existing_list[1:]:
                cognite_client.functions.schedules.delete(existing.id)
        schedule = existing_list[0]
    else:
        schedule = cognite_client.functions.schedules.create(
            name=name,
            cron_expression="0 7 * * MON",
            description="Original description.",
            function_external_id=dummy_function.external_id,
        )
    if schedule.function_external_id is None:
        schedule.function_external_id = dummy_function.external_id
    if schedule.function_id is not None:
        schedule.function_id = None
    return schedule


@pytest.fixture()
def raw_data() -> RowWriteList:
    return RowWriteList(
        [
            RowWrite(
                key=f"row{i}",
                columns={
                    "StringCol": f"value{i % 3}",
                    "IntegerCol": i % 5,
                    "BooleanCol": [True, False][i % 2],
                    "FloatCol": i * 0.1,
                    "EmptyCol": None,
                    "ArrayCol": [i, i + 1, i + 2] if i % 2 == 0 else None,
                    "ObjectCol": {"nested_key": f"nested_value_{i}" if i % 2 == 0 else None},
                },
            )
            for i in range(10)
        ]
    )


@pytest.fixture()
def populated_raw_table(toolkit_client: ToolkitClient, raw_data: RowWriteList) -> RawTable:
    db_name = "toolkit_test_db"
    table_name = "toolkit_test_profiling_table"
    existing_dbs = toolkit_client.raw.databases.list(limit=-1)
    existing_table_names: set[str] = set()
    if db_name in {db.name for db in existing_dbs}:
        existing_table_names = {table.name for table in toolkit_client.raw.tables.list(db_name=db_name, limit=-1)}
    if table_name not in existing_table_names:
        toolkit_client.raw.rows.insert(db_name, table_name, raw_data, ensure_parent=True)
    return RawTable(db_name, table_name)


@pytest.fixture(scope="session")
def aggregator_raw_db(toolkit_client: ToolkitClient) -> str:
    loader = RawDatabaseCRUD.create_loader(toolkit_client)
    db_name = "toolkit_aggregators_test_db"
    if not loader.retrieve([RawDatabase(db_name=db_name)]):
        loader.create(RawDatabaseList([RawDatabase(db_name=db_name)]))
    return db_name


@pytest.fixture(scope="session")
def aggregator_two_datasets(toolkit_client: ToolkitClient) -> DataSetList:
    datasets = DataSetWriteList(
        [
            DataSetWrite(external_id="toolkit_aggregators_test_dataset_1", name="Toolkit Aggregators Test Dataset 1"),
            DataSetWrite(external_id="toolkit_aggregators_test_dataset_2", name="Toolkit Aggregators Test Dataset 2"),
        ]
    )
    retrieved = toolkit_client.data_sets.retrieve_multiple(
        external_ids=datasets.as_external_ids(), ignore_unknown_ids=True
    )
    if not retrieved:
        return toolkit_client.data_sets.create(datasets)

    return retrieved


@pytest.fixture(scope="session")
def aggregator_root_asset(toolkit_client: ToolkitClient, aggregator_two_datasets: DataSetList) -> Asset:
    root_asset = AssetWrite(
        name="Toolkit Aggregators Test Root Asset",
        external_id="toolkit_aggregators_test_root_asset",
        data_set_id=aggregator_two_datasets[0].id,
    )
    retrieved = toolkit_client.assets.retrieve(external_id=root_asset.external_id)
    if retrieved is None:
        return toolkit_client.assets.create(root_asset)
    return retrieved


def create_raw_table_with_data(client: ToolkitClient, table: RawTable, rows: list[RowWrite]) -> None:
    loader = RawTableCRUD.create_loader(client)
    existing_tables = loader.retrieve([table])
    if not existing_tables:
        loader.create(RawTableList([table]))
    data = client.raw.rows.list(table.db_name, table.table_name, limit=len(rows))
    if not data:
        client.raw.rows.insert(table.db_name, table.table_name, rows)


def upsert_transformation_with_run(
    toolkit_client: ToolkitClient, transformation: TransformationWrite
) -> Transformation:
    retrieved = toolkit_client.transformations.retrieve(external_id=transformation.external_id)
    if retrieved is None:
        created = toolkit_client.transformations.create(transformation)
    else:
        created = retrieved
    if created.last_finished_job is None:
        nonce = toolkit_client.iam.sessions.create(session_type="ONESHOT_TOKEN_EXCHANGE")
        response = toolkit_client.post(
            url=f"/api/v1/projects/{toolkit_client.config.project}/transformations/run",
            json={
                "id": created.id,
                "nonce": {
                    "sessionId": nonce.id,
                    "nonce": nonce.nonce,
                    "cdfProjectName": toolkit_client.config.project,
                },
            },
        )
        job = TransformationJob._load(response.json(), cognite_client=toolkit_client)
        job.wait()
        assert job.error is None
    return created


@pytest.fixture(scope="session")
def aggregator_assets(
    toolkit_client: ToolkitClient, aggregator_raw_db: str, aggregator_root_asset: Asset
) -> Transformation:
    table_name = ASSET_TABLE
    rows = [
        RowWrite(
            key=f"asset_00{i}",
            columns={
                "name": f"Asset 00{i}",
                "externalId": f"asset_00{i}",
                "parentExternalId": aggregator_root_asset.external_id,
            },
        )
        for i in range(1, ASSET_COUNT - 1 + 1)  # -1 for root asset, +1 for inclusive range
    ]
    create_raw_table_with_data(
        toolkit_client,
        RawTable(db_name=aggregator_raw_db, table_name=table_name),
        rows,
    )

    transformation = TransformationWrite(
        external_id=ASSET_TRANSFORMATION,
        name="Toolkit Aggregators Test Asset Transformation",
        destination=TransformationDestination("assets"),
        query=f"""SELECT name as name, externalId as externalId, dataset_id('{ASSET_DATASET}') as dataSetId, parentExternalId as parentExternalId
FROM `{aggregator_raw_db}`.`{table_name}`""",
        ignore_null_fields=True,
    )
    created = upsert_transformation_with_run(toolkit_client, transformation)
    return created


@pytest.fixture(scope="session")
def aggregator_asset_list(
    toolkit_client: ToolkitClient,
    aggregator_root_asset: Asset,
    aggregator_assets: Transformation,
    aggregator_two_datasets: DataSetList,
) -> AssetList:
    return toolkit_client.assets.list(
        asset_subtree_ids=[aggregator_root_asset.id],
    )


@pytest.fixture(scope="session")
def aggregator_events(
    toolkit_client: ToolkitClient,
    aggregator_raw_db: str,
    aggregator_asset_list: AssetList,
    aggregator_two_datasets: DataSetList,
) -> Transformation:
    table_name = EVENT_TABLE
    assets = aggregator_asset_list
    rows = [
        RowWrite(
            key=f"event_00{i}",
            columns={
                "externalId": f"event_00{i}",
                "name": f"Event 00{i}",
                "startTime": 1000000000 + i * 1000,  # Staggered start times
                "endTime": 1000000000 + i * 1000 + 500,  # 500ms duration
                "assetIds": [assets[i % len(assets)].id],
            },
        )
        for i in range(1, EVENT_COUNT + 1)  # +1 for inclusive range
    ]
    create_raw_table_with_data(
        toolkit_client,
        RawTable(db_name=aggregator_raw_db, table_name=table_name),
        rows,
    )
    transformation = TransformationWrite(
        external_id=EVENT_TRANSFORMATION,
        name="Toolkit Aggregators Test Event Transformation",
        destination=TransformationDestination("events"),
        query=f"""SELECT externalId as externalId, name as name, timestamp_millis(startTime) as startTime, timestamp_millis(endTime) as endTime,
assetIds as assetIds, dataset_id('{EVENT_DATASET}') as dataSetId
FROM `{aggregator_raw_db}`.`{table_name}`""",
        ignore_null_fields=True,
    )
    created = upsert_transformation_with_run(toolkit_client, transformation)
    return created


@pytest.fixture(scope="session")
def aggregator_files(
    toolkit_client: ToolkitClient,
    aggregator_raw_db: str,
    aggregator_two_datasets: DataSetList,
    aggregator_asset_list: AssetList,
) -> Transformation:
    table_name = FILE_TABLE
    assets = aggregator_asset_list
    rows = [
        RowWrite(
            key=f"file_00{i}",
            columns={
                "externalId": f"file_00{i}",
                "name": f"File 00{i}",
                "assetIds": [assets[i % len(assets)].id],  # Assign to one of the assets
                "mimeType": "application/text",
            },
        )
        for i in range(1, FILE_COUNT + 1)  # +1 for inclusive range
    ]
    # all_files = toolkit_client.files.list(limit=-1)
    # to_delete = [file for file in all_files if file.external_id and file.external_id.startswith("file_00")]
    # if to_delete:
    #     toolkit_client.files.delete(id=[file.id for file in to_delete], ignore_unknown_ids=True)

    create_raw_table_with_data(
        toolkit_client,
        RawTable(db_name=aggregator_raw_db, table_name=table_name),
        rows,
    )
    transformation = TransformationWrite(
        external_id=FILE_TRANSFORMATION,
        name="Toolkit Aggregators Test File Transformation",
        destination=TransformationDestination("files"),
        query=f"""SELECT externalId as externalId, name as name, assetIds as assetIds,
dataset_id('{FILE_DATASET}') as dataSetId, mimeType as mimeType
FROM `{aggregator_raw_db}`.`{table_name}`""",
        ignore_null_fields=True,
    )
    created = upsert_transformation_with_run(toolkit_client, transformation)

    # Upload content for the files
    external_ids = [row.columns["externalId"] for row in rows]
    filemetadata = toolkit_client.files.retrieve_multiple(external_ids=external_ids)
    is_uploaded_by_external_id = {
        file.external_id: file.uploaded for file in filemetadata if file.external_id is not None
    }
    for row in rows:
        external_id = row.columns["externalId"]
        if is_uploaded_by_external_id.get(external_id):
            continue
        file_content = f"Content of {external_id}"
        toolkit_client.files.upload_content_bytes(file_content, external_id=external_id)

    return created


@pytest.fixture(scope="session")
def aggregator_time_series(
    toolkit_client: ToolkitClient,
    aggregator_raw_db: str,
    aggregator_two_datasets: DataSetList,
    aggregator_asset_list: AssetList,
) -> Transformation:
    table_name = TIMESERIES_TABLE
    assets = aggregator_asset_list
    rows = [
        RowWrite(
            key=f"timeseries_00{i}",
            columns={
                "externalId": f"timeseries_00{i}",
                "name": f"Time Series 00{i}",
                "assetId": assets[i % len(assets)].id,  # Assign to one of the assets
                "isString": False,
                "isStep": False,
            },
        )
        for i in range(1, TIMESERIES_COUNT + 1)  # +1 for inclusive range
    ]
    create_raw_table_with_data(
        toolkit_client,
        RawTable(db_name=aggregator_raw_db, table_name=table_name),
        rows,
    )
    transformation = TransformationWrite(
        external_id=TIMESERIES_TRANSFORMATION,
        name="Toolkit Aggregators Test Time Series Transformation",
        destination=TransformationDestination("timeseries"),
        query=f"""SELECT externalId as externalId, name as name, assetId as assetId, isString as isString, isStep as isStep,
dataset_id('{TIMESERIES_DATASET}') as dataSetId
FROM `{aggregator_raw_db}`.`{table_name}`""",
        ignore_null_fields=True,
    )
    created = upsert_transformation_with_run(toolkit_client, transformation)
    return created


@pytest.fixture(scope="session")
def aggregator_sequences(
    toolkit_client: ToolkitClient,
    aggregator_raw_db: str,
    aggregator_two_datasets: DataSetList,
    aggregator_asset_list: AssetList,
) -> Transformation:
    table_name = SEQUENCE_TABLE
    assets = aggregator_asset_list
    rows = [
        RowWrite(
            key=f"sequence_00{i}",
            columns={
                "externalId": f"sequence_00{i}",
                "name": f"Sequence 00{i}",
                "assetId": assets[i % len(assets)].id,  # Assign to one of the assets
                "columns": [
                    {"name": f"Column {j}", "valueType": "STRING", "externalId": f"column_{j}"} for j in range(1, 4)
                ],
            },
        )
        for i in range(1, SEQUENCE_COUNT + 1)  # +1 for inclusive range
    ]
    create_raw_table_with_data(
        toolkit_client,
        RawTable(db_name=aggregator_raw_db, table_name=table_name),
        rows,
    )
    transformation = TransformationWrite(
        external_id=SEQUENCE_TRANSFORMATION,
        name="Toolkit Aggregators Test Sequence Transformation",
        destination=TransformationDestination("sequences"),
        query=f"""SELECT externalId as externalId, name as name, assetId as assetId, columns as columns, dataset_id('{SEQUENCE_DATASET}') as dataSetId
FROM `{aggregator_raw_db}`.`{table_name}`""",
        ignore_null_fields=True,
    )
    created = upsert_transformation_with_run(toolkit_client, transformation)
    return created


@pytest.fixture()
def disable_throttler(
    toolkit_client: ToolkitClient,
) -> None:
    def no_op(*args, **kwargs) -> None:
        """No operation function to replace the write_last_call_epoc function."""
        pass

    always_enabled = MagicMock(spec=ThrottlerState)
    # We mock the TrottlerState the mock object will always pass the throttling check.
    always_enabled.get.return_value = MagicMock(spec=ThrottlerState)
    with (
        patch(f"{raw_row_count.__module__}.ThrottlerState", always_enabled),
    ):
        yield


@dataclass
class HierarchyMinimal:
    root_asset: Asset
    child_asset: Asset
    event: Event
    file: FileMetadata
    timeseries: TimeSeries
    dataset: DataSet


@pytest.fixture(scope="session")
def migration_hierarchy_minimal(toolkit_client: ToolkitClient) -> HierarchyMinimal:
    root = "migration_test_root_asset"
    client = toolkit_client
    dataset_write = DataSetWrite(
        external_id="migration_test_dataset",
        name="Migration Test DataSet",
        description="DataSet for migration integration tests",
    )
    data_set = client.data_sets.retrieve(external_id=dataset_write.external_id)
    if data_set is None:
        data_set = client.data_sets.create(dataset_write)
    asset_source = "ToolkitAsset"
    event_source = "ToolkitEvent"
    file_source = "ToolkitFile"
    assets = AssetWriteList(
        [
            AssetWrite(
                name="Migration Test Root Asset",
                external_id=root,
                description="Root asset for migration integration tests",
                data_set_id=data_set.id,
                source=asset_source,
            ),
            AssetWrite(
                name="Migration Test Child Asset 1",
                external_id="migration_test_child_asset_1",
                description="Child asset 1 for migration integration tests",
                parent_external_id=root,
                data_set_id=data_set.id,
                source=asset_source,
            ),
        ]
    )
    created_assets = client.assets.upsert(assets, mode="replace")
    child_asset = created_assets[1]
    event = EventWrite(
        external_id="migration_test_event",
        data_set_id=data_set.id,
        start_time=1_600_000_000_000,
        end_time=1_600_000_000_500,
        type="WorkOrder",
        asset_ids=[child_asset.id],
        source=event_source,
    )
    created_event = client.events.upsert(event, mode="replace")
    file = FileMetadataWrite(
        external_id="migration_test_file",
        name="migration_test_file.txt",
        mime_type="text/plain",
        data_set_id=data_set.id,
        asset_ids=[child_asset.id],
        source=file_source,
    )
    created_file = client.files.retrieve(external_id=file.external_id)
    if created_file is None:
        created_file, _ = client.files.create(file, overwrite=True)
    if not created_file.uploaded:
        client.files.upload_content_bytes("This is a test file.", external_id=created_file.external_id)

    timeseries = TimeSeriesWrite(
        name="Migration Test Time Series",
        external_id="migration_test_timeseries",
        unit="C",
        is_step=False,
        is_string=False,
        asset_id=child_asset.id,
        data_set_id=data_set.id,
        unit_external_id="temperature:deg_c",
    )
    created_timeseries = client.time_series.retrieve(external_id=timeseries.external_id)
    if created_timeseries is None:
        created_timeseries = client.time_series.create(timeseries)

    if not client.time_series.data.retrieve_latest(external_id=timeseries.external_id):
        client.time_series.data.insert(
            external_id=timeseries.external_id,
            datapoints=[(1_600_000_000_000, 20.0), (1_600_000_000_500, 21.5)],
        )

    return HierarchyMinimal(
        root_asset=created_assets[0],
        child_asset=child_asset,
        event=created_event,
        file=created_file,
        timeseries=created_timeseries,
        dataset=data_set,
    )


def test_migration_hierarchy(migration_hierarchy_minimal: HierarchyMinimal) -> None:
    assert True, "Fixture migration_hierarchy_minimal failed"



import re
import sys
from types import MappingProxyType
from typing import Any, ClassVar, Literal, cast

from pydantic import Field, ModelWrapValidatorHandler, model_serializer, model_validator
from pydantic_core.core_schema import SerializerFunctionWrapHandler

from cognite_toolkit._cdf_tk.constants import (
    CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN,
    DM_EXTERNAL_ID_PATTERN,
    DM_VERSION_PATTERN,
    SPACE_FORMAT_PATTERN,
)
from cognite_toolkit._cdf_tk.utils.collection import humanize_collection

from .base import BaseModelResource
from .container_field_definitions import ContainerReference

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

KEY_PATTERN = re.compile(CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN)


class ViewReference(BaseModelResource, populate_by_name=True):
    type: Literal["view"] = "view"
    space: str = Field(
        description="Id of the space that the view belongs to.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the view.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    version: str = Field(
        description="Version of the view.",
        max_length=43,
        pattern=DM_VERSION_PATTERN,
    )


class DirectRelationReference(BaseModelResource):
    space: str = Field(
        description="Id of the space that the instance belongs to.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the instance.",
        min_length=1,
        max_length=255,
    )


class ThroughRelationReference(BaseModelResource):
    source: ViewReference | ContainerReference = Field(
        description="Reference to the view or container from where this relation is inherited.",
    )
    identifier: str = Field(
        description="Identifier of the relation in the source view or container.",
        min_length=1,
        max_length=255,
    )


class ViewProperty(BaseModelResource):
    name: str | None = Field(
        default=None,
        description="Name of the property.",
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="Description of the content and suggested use for this property..",
        max_length=1024,
    )

    @model_validator(mode="wrap")
    @classmethod
    def find_property_type_cls(cls, data: Any, handler: ModelWrapValidatorHandler[Self]) -> Self:
        if isinstance(data, ViewProperty):
            return cast(Self, data)
        if not isinstance(data, dict):
            raise ValueError(f"Invalid property type data '{type(data)}' expected dict")

        if cls is not ViewProperty:
            data_copy = dict(data)
            if cls in _CONNECTION_DEFINITION_CLASS_BY_TYPE.values():
                data_copy.pop("connectionType", None)
            return handler(data_copy)

        cls_: type[ContainerViewProperty] | type[ConnectionDefinition]
        if "container" in data:
            cls_ = ContainerViewProperty
        elif "connectionType" in data:
            connection_type = data.get("connectionType")
            if connection_type is None:
                raise ValueError("Missing 'connectionType' field in connection definition data")
            if connection_type not in _CONNECTION_DEFINITION_CLASS_BY_TYPE:
                raise ValueError(
                    f"invalid connection type '{connection_type}'. Expected one of {humanize_collection(_CONNECTION_DEFINITION_CLASS_BY_TYPE.keys(), bind_word='or')}"
                )
            cls_ = _CONNECTION_DEFINITION_CLASS_BY_TYPE[connection_type]
        else:
            raise ValueError(
                "Invalid Property data. If it is a connection definition, it must contain 'connectionType' field. If it is a view property, it must contain 'container' and 'containerPropertIdentifier' field."
            )

        data_copy = dict(data)
        data_copy.pop("connectionType", None)
        return cast(Self, cls_.model_validate(data_copy))

    @model_serializer(mode="wrap", when_used="always", return_type=dict)
    def serialize_property_type(self, handler: SerializerFunctionWrapHandler) -> dict:
        serialized_data = handler(self)
        if hasattr(self, "connection_type"):
            serialized_data["connectionType"] = self.__class__.connection_type
        return serialized_data


class ContainerViewProperty(ViewProperty):
    container: ContainerReference = Field(
        description="Reference to the container where this property is defined.",
    )
    container_property_identifier: str = Field(
        description="Identifier of the property in the container.",
        min_length=1,
        max_length=255,
        pattern=CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN,
    )
    source: ViewReference | None = Field(
        default=None,
        description="Indicates on what type a referenced direct relation is expected to be. Only applicable for direct relation properties.",
    )


class ConnectionDefinition(ViewProperty):
    connection_type: ClassVar[str]
    source: ViewReference = Field(
        description="Indicates the view which is either the target node(s) or the node(s) containing the direct relation property."
    )


class EdgeConnectionDefinition(ConnectionDefinition):
    connection_type: ClassVar[Literal["single_edge_connection", "multi_edge_connection"]]
    type: DirectRelationReference = Field(
        description="Reference to the node pointed to by the direct relation.",
    )
    edge_source: ViewReference | None = Field(
        default=None,
        description="Reference to the view from where this edge connection is inherited.",
    )
    direction: Literal["outwards", "inwards"]


class SingleEdgeConnectionDefinition(EdgeConnectionDefinition):
    connection_type = "single_edge_connection"


class MultiEdgeConnectionDefinition(EdgeConnectionDefinition):
    connection_type = "multi_edge_connection"


class ReverseDirectRelationConnectionDefinition(ConnectionDefinition):
    connection_type: ClassVar[Literal["single_reverse_direct_relation", "multi_reverse_direct_relation"]]
    through: ThroughRelationReference = Field(
        description="The view or container of the node containing the direct relation property.",
    )


class SingleReverseDirectRelationConnectionDefinition(ReverseDirectRelationConnectionDefinition):
    connection_type = "single_reverse_direct_relation"


class MultiReverseDirectRelationConnectionDefinition(ReverseDirectRelationConnectionDefinition):
    connection_type = "multi_reverse_direct_relation"


def get_connection_definition_type_leaf_classes(base_class: type[ConnectionDefinition]) -> list:
    subclasses = base_class.__subclasses__()
    result = []

    if not subclasses:
        if base_class is not ConnectionDefinition:
            result.append(base_class)
    else:
        for subclass in subclasses:
            result.extend(get_connection_definition_type_leaf_classes(subclass))

    return result


_CONNECTION_DEFINITION_CLASS_BY_TYPE: MappingProxyType[str, type[ConnectionDefinition]] = MappingProxyType(
    {
        c.connection_type: c
        for c in get_connection_definition_type_leaf_classes(ConnectionDefinition)
        if hasattr(c, "connection_type") and c.connection_type is not None
    }
)



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.capabilities import Capability
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic


def all_acls() -> Iterable:
    acl_list = [
        {"agentsAcl": {"actions": ["READ", "WRITE", "RUN"], "scope": {"all": {}}}},
        {"annotationsAcl": {"actions": ["WRITE", "READ", "SUGGEST", "REVIEW"], "scope": {"all": {}}}},
        {"appConfigAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"appConfigAcl": {"actions": ["READ", "WRITE"], "scope": {"appScope": {"apps": ["SEARCH"]}}}},
        {"assetsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"assetsAcl": {"actions": ["READ", "WRITE"], "scope": {"datasetScope": {"ids": ["myDataSet"]}}}},
        {"auditlogAcl": {"actions": ["READ"], "scope": {"all": {}}}},
        {"dataModelInstancesAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"dataModelInstancesAcl": {"actions": ["READ"], "scope": {"spaceIdScope": {"spaceIds": ["maintain"]}}}},
        {
            "dataModelInstancesAcl": {
                "actions": ["WRITE_PROPERTIES"],
                "scope": {"spaceIdScope": {"spaceIds": ["tech-space"]}},
            }
        },
        {"dataModelsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {
            "dataModelsAcl": {
                "actions": ["READ"],
                "scope": {"spaceIdScope": {"spaceIds": ["maintain", "main-data"]}},
            }
        },
        {"datasetsAcl": {"actions": ["READ", "WRITE", "OWNER"], "scope": {"all": {}}}},
        {"datasetsAcl": {"actions": ["READ", "WRITE", "OWNER"], "scope": {"idScope": {"ids": ["my_dataset"]}}}},
        {"diagramParsingAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"digitalTwinAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"documentFeedbackAcl": {"actions": ["CREATE", "READ", "DELETE"], "scope": {"all": {}}}},
        {"documentPipelinesAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"entitymatchingAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"eventsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"eventsAcl": {"actions": ["READ", "WRITE"], "scope": {"datasetScope": {"ids": ["myDataSet"]}}}},
        {"eventsAcl": {"actions": ["READ"], "scope": {"datasetScope": {"ids": ["myDataSet", "myDataSet2"]}}}},
        {"extractionConfigsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"extractionPipelinesAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {
            "extractionPipelinesAcl": {
                "actions": ["READ", "WRITE"],
                "scope": {"idScope": {"ids": ["myPipeline", "myPipeline"]}},
            }
        },
        {
            "extractionPipelinesAcl": {
                "actions": ["READ", "WRITE"],
                "scope": {"datasetScope": {"ids": ["myDataSet", "myDataSet2"]}},
            }
        },
        {"extractionRunsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"filePipelinesAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"filesAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"filesAcl": {"actions": ["READ", "WRITE"], "scope": {"datasetScope": {"ids": ["myDataSet", "myDataSet2"]}}}},
        {"functionsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"genericsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"groupsAcl": {"actions": ["LIST", "READ", "DELETE", "UPDATE", "CREATE"], "scope": {"all": {}}}},
        {"groupsAcl": {"actions": ["READ", "CREATE", "UPDATE", "DELETE"], "scope": {"currentuserscope": {}}}},
        {"hostedExtractorsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"labelsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"locationFiltersAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {
            "locationFiltersAcl": {
                "actions": ["READ", "WRITE"],
                "scope": {"idScope": {"ids": ["dataset", "otherDataset"]}},
            }
        },
        {"modelHostingAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"monitoringTasksAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"notificationsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"pipelinesAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {
            "postgresGatewayAcl": {
                "actions": ["READ", "WRITE"],
                "scope": {"usersScope": {"usernames": ["test_1", "test_2"]}},
            }
        },
        {"projectsAcl": {"actions": ["UPDATE", "LIST", "READ", "CREATE", "DELETE"], "scope": {"all": {}}}},
        {"rawAcl": {"actions": ["READ", "WRITE", "LIST"], "scope": {"all": {}}}},
        {
            "rawAcl": {
                "actions": ["READ", "WRITE", "LIST"],
                "scope": {"tableScope": {"dbsToTables": {"no table in this": []}}},
            }
        },
        {
            "rawAcl": {
                "actions": ["READ", "WRITE", "LIST"],
                "scope": {"tableScope": {"dbsToTables": {"test db 1": ["empty tbl", "test tbl 1"]}}},
            }
        },
        {"relationshipsAcl": {"actions": ["READ"], "scope": {"all": {}}}},
        {"relationshipsAcl": {"actions": ["READ"], "scope": {"datasetScope": {"ids": ["myDataSet", "myDataSet2"]}}}},
        {"roboticsAcl": {"actions": ["READ", "CREATE", "UPDATE", "DELETE"], "scope": {"all": {}}}},
        {"roboticsAcl": {"actions": ["READ"], "scope": {"datasetScope": {"ids": ["myDataSet"]}}}},
        {"sapWritebackAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"sapWritebackAcl": {"actions": ["READ", "WRITE"], "scope": {"instancesScope": {"instances": ["123", "456"]}}}},
        {"sapWritebackRequestsAcl": {"actions": ["WRITE", "LIST"], "scope": {"all": {}}}},
        {
            "sapWritebackRequestsAcl": {
                "actions": ["WRITE", "LIST"],
                "scope": {"instancesScope": {"instances": ["123", "456"]}},
            }
        },
        {"scheduledCalculationsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {
            "securityCategoriesAcl": {
                "actions": ["DELETE", "MEMBEROF", "LIST", "CREATE", "UPDATE"],
                "scope": {"all": {}},
            }
        },
        {
            "securityCategoriesAcl": {
                "actions": ["MEMBEROF", "LIST", "CREATE", "UPDATE", "DELETE"],
                "scope": {"idscope": {"ids": ["myCategory", "myOtherCategory"]}},
            }
        },
        {"seismicAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"seismicAcl": {"actions": ["WRITE"], "scope": {"partition": {"partitionIds": [123, 456]}}}},
        {"sequencesAcl": {"actions": ["READ"], "scope": {"all": {}}}},
        {"sequencesAcl": {"actions": ["WRITE"], "scope": {"datasetScope": {"ids": ["myDataSet", "myDataSet2"]}}}},
        {"sessionsAcl": {"actions": ["LIST", "CREATE", "DELETE"], "scope": {"all": {}}}},
        {"templateGroupsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {
            "templateGroupsAcl": {
                "actions": ["READ", "WRITE"],
                "scope": {"datasetScope": {"ids": ["myDataSet", "myDataSet2"]}},
            }
        },
        {
            "templateInstancesAcl": {
                "actions": ["READ", "WRITE"],
                "scope": {"datasetScope": {"ids": ["myDataSet", "myDataSet2"]}},
            }
        },
        {"templateInstancesAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"threedAcl": {"actions": ["READ", "CREATE", "UPDATE", "DELETE"], "scope": {"all": {}}}},
        {"timeSeriesAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {
            "timeSeriesAcl": {
                "actions": ["READ", "WRITE"],
                "scope": {"datasetScope": {"ids": ["myDataSet", "myDataSet2"]}},
            }
        },
        {"timeSeriesAcl": {"actions": ["READ"], "scope": {"idscope": {"ids": ["myTimeseries", "myOtherTimeseries"]}}}},
        {"timeSeriesAcl": {"actions": ["WRITE", "READ"], "scope": {"assetRootIdScope": {"rootIds": ["myAsset"]}}}},
        {"transformationsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"transformationsAcl": {"actions": ["READ", "WRITE"], "scope": {"datasetScope": {"ids": ["myDataSet"]}}}},
        {"visionModelAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"wellsAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {"workflowOrchestrationAcl": {"actions": ["READ", "WRITE"], "scope": {"all": {}}}},
        {
            "workflowOrchestrationAcl": {
                "actions": ["READ", "WRITE"],
                "scope": {"datasetScope": {"ids": ["myDataSet", "myDataSet2"]}},
            }
        },
    ]

    yield from (pytest.param(acl, id=next(iter(acl.keys()))) for acl in acl_list)


class TestCapabilities:
    @pytest.mark.parametrize("acl", all_acls())
    def test_load_dump_capability(self, acl: dict[str, object]) -> None:
        capability = Capability.model_validate(acl)
        assert capability.model_dump(by_alias=True, exclude_unset=True) == acl

    @pytest.mark.parametrize(
        "data, expected_errors",
        [
            pytest.param(
                {"datasetsAcl": {"actions": ["READ", "WRITE", "OWNER"], "scope": {"idscope": {"ids": ["my_dataset"]}}}},
                ["In field scope invalid scope name 'idscope'. Expected all or idScope"],
                id="Wrong case for datasetsAcl idScope",
            ),
            pytest.param(
                {
                    "extractionPipelinesAcl": {
                        "actions": ["OWNER", "READ"],
                        "scope": {"idscope": {"ids": ["my_pipeline"]}},
                    }
                },
                [
                    "In field scope invalid scope name 'idscope'. Expected all, datasetScope or idScope",
                    "In actions[1] input should be 'READ' or 'WRITE'. Got 'OWNER'.",
                ],
                id="Wrong case for extractionPipelinesAcl idScope",
            ),
            pytest.param(
                {"appConfigAcl": {"actions": ["READ", "WRITE"], "scope": {"appScope": {"apps": ["UNKNOWN_APP"]}}}},
                ["In scope.apps[1] input should be 'SEARCH'. Got 'UNKNOWN_APP'."],
                id="Invalid app name in appConfigAcl",
            ),
            pytest.param(
                {"agentsAcl": {"actions": ["READ", "LIST"], "scope": {"idScope": {"ids": ["my_agent"]}}}},
                [
                    "In field scope invalid scope name 'idScope'. Expected all",
                    "In actions[2] input should be 'READ', 'WRITE' or 'RUN'. Got 'LIST'.",
                ],
                id="AgentsAcl with invalid scope and action",
            ),
        ],
    )
    def test_invalid_capability(self, data: dict[str, object], expected_errors: list[str]) -> None:
        warnings = validate_resource_yaml_pydantic(data, Capability, source_file=Path("filename.yaml"))
        assert len(warnings) == 1
        warning = warnings[0]
        assert isinstance(warning, ResourceFormatWarning)
        assert list(warning.errors) == expected_errors



from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.robotics.map import RobotMapYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


class TestRobotMapYAML:
    @pytest.mark.parametrize("data", list(find_resources("Map", "robotics")))
    def test_load_valid_robot_map_file(self, data: dict[str, object]) -> None:
        loaded = RobotMapYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize(
        "data",
        [
            {"name": "Warehouse Map", "externalId": "map_001", "mapType": "WAYPOINTMAP"},
            {
                "name": "3D Model Map",
                "externalId": "map_3d_001",
                "mapType": "THREEDMODEL",
                "description": "3D model of the facility",
                "frameExternalId": "frame_001",
                "data": {"modelUrl": "https://example.com/model.obj"},
                "locationExternalId": "location_001",
                "scale": 1.0,
            },
        ],
    )
    def test_load_valid_map(self, data: dict[str, object]) -> None:
        """Test valid map configurations"""
        loaded = RobotMapYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize(
        "data,expected_error",
        [
            ({"name": "Missing Map Type", "externalId": "map_001"}, {"Missing required field: 'mapType'"}),
            (
                {"name": "Invalid Map", "externalId": "map_invalid", "mapType": "INVALID_TYPE"},
                {
                    "In field mapType input should be 'WAYPOINTMAP', 'THREEDMODEL', 'TWODMAP' or 'POINTCLOUD'. Got 'INVALID_TYPE'."
                },
            ),
            (
                {
                    "name": "3D Model Map",
                    "externalId": "map_3d_001",
                    "mapType": "THREEDMODEL",
                    "description": "3D model of the facility",
                    "frameExternalId": "frame_001",
                    "data": {"modelUrl": "https://example.com/model.obj"},
                    "locationExternalId": "location_001",
                    "scale": 1.1,
                },
                {"In field scale input should be less than or equal to 1"},
            ),
        ],
    )
    def test_invalid_map_configurations(self, data: dict[str, object], expected_error: str) -> None:
        """Test invalid map configurations"""
        warning_list = validate_resource_yaml_pydantic(data, RobotMapYAML, Path("test.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert set(format_warning.errors) == expected_error



from __future__ import annotations

from pathlib import Path

import yaml

from cognite_toolkit._cdf_tk.data_classes import BuildVariables, ModuleLocation


class TestBuildVariables:
    def test_replace_preserve_data_type(self) -> None:
        source_yaml = """text: {{ my_text }}
bool: {{ my_bool }}
integer: {{ my_integer }}
float: {{ my_float }}
digit_string: {{ my_digit_string }}
quoted_string: "{{ my_quoted_string }}"
list: {{ my_list }}
null_value: {{ my_null_value }}
single_quoted_string: '{{ my_single_quoted_string }}'
composite: 'some_prefix_{{ my_composite }}'
prefix_text: {{ my_prefix_text }}
suffix_text: {{ my_suffix_text }}
"""
        variables = BuildVariables.load_raw(
            {
                "my_text": "some text",
                "my_bool": True,
                "my_integer": 123,
                "my_float": 123.456,
                "my_digit_string": "123",
                "my_quoted_string": "456",
                "my_list": ["one", "two", "three"],
                "my_null_value": None,
                "my_single_quoted_string": "789",
                "my_composite": "the suffix",
                "my_prefix_text": "prefix:",
                "my_suffix_text": ":suffix",
            },
            available_modules=set(),
            selected_modules=set(),
        )

        result = variables.replace(source_yaml)

        loaded = yaml.safe_load(result)
        assert loaded == {
            "text": "some text",
            "bool": True,
            "integer": 123,
            "float": 123.456,
            "digit_string": "123",
            "quoted_string": "456",
            "list": ["one", "two", "three"],
            "null_value": None,
            "single_quoted_string": "789",
            "composite": "some_prefix_the suffix",
            "prefix_text": "prefix:",
            "suffix_text": ":suffix",
        }

    def test_replace_not_preserve_type(self) -> None:
        source_yaml = """dataset_id('{{dataset_external_id}}')"""
        variables = BuildVariables.load_raw(
            {
                "dataset_external_id": "ds_external_id",
            },
            available_modules=set(),
            selected_modules=set(),
        )

        result = variables.replace(source_yaml, Path("test.sql"))

        assert result == "dataset_id('ds_external_id')"

    def test_replace_sql_list(self) -> None:
        """Test that lists with mixed types in SQL files are formatted correctly."""
        source_sql = """SELECT * FROM table WHERE column IN {{ my_list }}"""
        variables = BuildVariables.load_raw(
            {
                "my_list": ["A", 123, None, True],
            },
            available_modules=set(),
            selected_modules=set(),
        )

        result = variables.replace(source_sql, Path("test.sql"))

        assert result == "SELECT * FROM table WHERE column IN ('A', 123, NULL, True)"

    def test_replace_inline_sql_preserve_double_quotes(self) -> None:
        source_yaml = """externalId: some_id
name: Some Transformation
destination:
  type: nodes
  view:
    space: cdf_cdm
    externalId: CogniteTimeSeries
    version: v1
  instanceSpace: my_instance_space
query: >-
  select "fpso_{{location_id}}" as externalId, "{{location_ID}}" as uid, "{{location_ID}}" as description
"""
        variables = BuildVariables.load_raw(
            {
                "location_id": "uny",
                "location_ID": "UNY",
            },
            available_modules=set(),
            selected_modules=set(),
        )

        result = variables.replace(source_yaml, Path("test.Transformation.yaml"))

        loaded = yaml.safe_load(result)

        assert loaded["query"] == 'select "fpso_uny" as externalId, "UNY" as uid, "UNY" as description'

    def test_replace_yaml_query_field_list(self) -> None:
        """Test that lists in multiline query fields are converted to SQL-style tuples."""
        source_yaml = """externalId: some_id
name: Some Transformation
query: |
  SELECT * FROM table
  WHERE column IN {{ my_list }}
  AND other_column = 'value'
"""
        variables = BuildVariables.load_raw(
            {
                "my_list": ["X", "Y", "Z"],
            },
            available_modules=set(),
            selected_modules=set(),
        )

        result = variables.replace(source_yaml, Path("test.Transformation.yaml"))

        loaded = yaml.safe_load(result)
        assert "('X', 'Y', 'Z')" in loaded["query"]

    def test_replace_yaml_query_field_list_mixed_types(self) -> None:
        """Test that lists with mixed types in query fields are formatted correctly."""
        source_yaml = """externalId: some_id
name: Some Transformation
query: >-
  SELECT * FROM table WHERE column IN {{ my_list }}
"""
        variables = BuildVariables.load_raw(
            {
                "my_list": ["A", 123, None, True],
            },
            available_modules=set(),
            selected_modules=set(),
        )

        result = variables.replace(source_yaml, Path("test.Transformation.yaml"))

        loaded = yaml.safe_load(result)
        assert loaded["query"] == "SELECT * FROM table WHERE column IN ('A', 123, NULL, True)"

    def test_replace_yaml_non_query_field_preserves_list(self) -> None:
        """Test that lists outside query fields still use YAML list format."""
        source_yaml = """externalId: some_id
name: Some Transformation
tags: {{ my_list }}
query: >-
  SELECT * FROM table
"""
        variables = BuildVariables.load_raw(
            {
                "my_list": ["tag1", "tag2"],
            },
            available_modules=set(),
            selected_modules=set(),
        )

        result = variables.replace(source_yaml)

        loaded = yaml.safe_load(result)
        # Tags should be a YAML list, not SQL tuple
        assert loaded["tags"] == ["tag1", "tag2"]

    def test_get_module_variables_variable_preference_order(self) -> None:
        source_yaml = """
modules:
  industry_apps:
      module_version: '1'
      pause_transformations: true
      apm_datamodel_space: APM_SourceData
      apm_sourcedata_model_version: '1.2.1'

      industry_apps_crna_common:
        apm_sourcedata_model_version: '1'
"""
        selected = {
            Path("."),
            Path("modules"),
            Path("modules/industry_apps"),
            Path("modules/industry_apps/industry_apps_crna_common"),
        }

        variables = BuildVariables.load_raw(
            yaml.safe_load(source_yaml), available_modules=selected, selected_modules=selected
        )

        assert len(variables) == 5
        location = ModuleLocation(
            Path("modules/industry_apps/industry_apps_crna_common"), Path("."), source_paths=[], is_selected=True
        )
        local_variables = variables.get_module_variables(location)[0]

        assert len(local_variables) == 4
        apm_sourcedata_model_version = next(
            (variable for variable in local_variables if variable.key == "apm_sourcedata_model_version"), None
        )
        assert apm_sourcedata_model_version.value == "1"



from ._client import HTTPClient
from ._data_classes import (
    DataBodyRequest,
    ErrorDetails,
    FailedRequestItems,
    FailedRequestMessage,
    FailedResponse,
    FailedResponseItems,
    HTTPMessage,
    ItemMessage,
    ItemsRequest,
    ParamRequest,
    RequestMessage,
    ResponseList,
    ResponseMessage,
    SimpleBodyRequest,
    SuccessResponse,
    SuccessResponseItems,
)
from ._exception import ToolkitAPIError

__all__ = [
    "DataBodyRequest",
    "ErrorDetails",
    "FailedRequestItems",
    "FailedRequestMessage",
    "FailedResponse",
    "FailedResponseItems",
    "HTTPClient",
    "HTTPMessage",
    "ItemMessage",
    "ItemsRequest",
    "ParamRequest",
    "RequestMessage",
    "ResponseList",
    "ResponseMessage",
    "SimpleBodyRequest",
    "SuccessResponse",
    "SuccessResponseItems",
    "ToolkitAPIError",
]



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.search_config import SearchConfigYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_search_config_test_cases() -> Iterable:
    yield pytest.param(
        {"useAsName": "name_prop"},
        {"Missing required field: 'view'"},
        id="Missing required field: view",
    )
    yield pytest.param(
        {
            "view": {"space": "my_space", "externalId": "my_view"},
            "columnsLayout": [{"property": 123}],  # property should be str
        },
        {
            "In columnsLayout[1].property input should be a valid string. Got 123 of type int. "
            "Hint: Use double quotes to force string.",
        },
        id="Invalid type in columns_layout property",
    )
    yield pytest.param(
        {
            "view": {"space": "my_space", "externalId": "my_view"},
            "unknown_field": "value",
        },
        {"Unused field: 'unknown_field'"},
        id="Unused field",
    )


class TestSearchConfigYAML:
    @pytest.mark.parametrize("data", list(find_resources("SearchConfig")))
    def test_load_valid_search_config(self, data: dict[str, object]) -> None:
        loaded = SearchConfigYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_search_config_test_cases()))
    def test_invalid_search_config_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, SearchConfigYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



import csv
import importlib.util
import json
import sys
import threading
from abc import ABC, abstractmethod
from collections import Counter
from collections.abc import Iterable, Mapping, Sequence
from datetime import date, datetime, timezone
from functools import lru_cache
from io import IOBase, TextIOWrapper
from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, Generic

import yaml

from cognite_toolkit._cdf_tk.exceptions import ToolkitMissingDependencyError, ToolkitTypeError, ToolkitValueError
from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses
from cognite_toolkit._cdf_tk.utils.collection import humanize_collection
from cognite_toolkit._cdf_tk.utils.file import sanitize_filename
from cognite_toolkit._cdf_tk.utils.table_writers import DataType

from ._base import T_IO, CellValue, Chunk, FileIO, SchemaColumn
from ._compression import Compression, Uncompressed

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


if TYPE_CHECKING:
    import pyarrow as pa
    import pyarrow.parquet as pq


class FileWriter(FileIO, ABC, Generic[T_IO]):
    def __init__(
        self, output_dir: Path, kind: str, compression: type[Compression], max_file_size_bytes: int = 128 * 1024 * 1024
    ) -> None:
        self.output_dir = output_dir
        self.kind = kind
        self.compression_cls = compression
        self.max_file_size_bytes = max_file_size_bytes
        self._file_count_by_filename: dict[str, int] = Counter()
        self._writer_by_filepath: dict[Path, T_IO] = {}
        self._lock = threading.Lock()

    @property
    def file_count(self) -> int:
        """Get the total number of files written."""
        with self._lock:
            return len(self._writer_by_filepath)

    def write_chunks(self, chunks: Iterable[Chunk], filestem: str = "") -> None:
        with self._lock:
            filepath = self._get_filepath(filestem)
            writer = self._get_writer(filepath, filestem)
            self._write(writer, chunks)

    def _get_filepath(self, filename: str) -> Path:
        # This method is now called within the lock context from write_chunks
        sanitized_name = f"{sanitize_filename(filename)}-" if filename else ""
        file_count = self._file_count_by_filename[filename]
        file_path = (
            self.output_dir
            / f"{sanitized_name}part-{file_count:04}.{self.kind}{self.format}{self.compression_cls.file_suffix}"
        )
        file_path.parent.mkdir(parents=True, exist_ok=True)
        return file_path

    def _get_writer(self, filepath: Path, filename_base: str) -> T_IO:
        # This method is now called within the lock context from write_chunks
        if filepath not in self._writer_by_filepath:
            self._writer_by_filepath[filepath] = self._create_writer(filepath)
        elif self._is_above_file_size_limit(filepath, self._writer_by_filepath[filepath]):
            self._writer_by_filepath[filepath].close()
            del self._writer_by_filepath[filepath]
            self._file_count_by_filename[filename_base] += 1
            new_filepath = self._get_filepath(filename_base)
            return self._get_writer(new_filepath, filename_base)
        return self._writer_by_filepath[filepath]

    def __enter__(self) -> Self:
        # Defensive check - should never trigger in normal usage
        if self._writer_by_filepath or self._file_count_by_filename:
            raise RuntimeError(f"{type(self).__name__} context manager should not be reused")
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        with self._lock:
            for writer in self._writer_by_filepath.values():
                writer.close()
            self._writer_by_filepath.clear()
            self._file_count_by_filename.clear()
            return None

    def _is_above_file_size_limit(self, filepath: Path, writer: T_IO) -> bool:
        """Check if the file size is above the limit."""
        try:
            writer.flush()
        except (AttributeError, ValueError):
            # Some writers might not support flush (e.g. already closed).
            # We can ignore this and proceed to check the file size on disk.
            pass
        return filepath.exists() and filepath.stat().st_size > self.max_file_size_bytes

    @abstractmethod
    def _create_writer(self, filepath: Path) -> T_IO:
        """Create a writer for the given file path."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    @abstractmethod
    def _write(self, writer: T_IO, chunks: Iterable[Chunk]) -> None:
        """Write the chunk to the file."""
        raise NotImplementedError("This method should be implemented in subclasses.")

    @classmethod
    def create_from_format(
        cls,
        format: str,
        output_dir: Path,
        kind: str,
        compression: type[Compression] = Uncompressed,
        columns: Sequence[SchemaColumn] | None = None,
    ) -> "FileWriter":
        if format not in FILE_WRITE_CLS_BY_FORMAT:
            raise ToolkitValueError(
                f"Unknown file format: {format}. Available formats: {humanize_collection(FILE_WRITE_CLS_BY_FORMAT.keys())}."
            )
        file_writs_cls = FILE_WRITE_CLS_BY_FORMAT[format]
        if issubclass(file_writs_cls, TableWriter) and columns is not None:
            return file_writs_cls(output_dir=output_dir, kind=kind, compression=compression, columns=columns)
        elif issubclass(file_writs_cls, TableWriter):
            raise ToolkitValueError(f"File format {format} requires columns to be provided for table writers.")
        else:
            return file_writs_cls(output_dir=output_dir, kind=kind, compression=compression)


class TableWriter(FileWriter[T_IO], ABC):
    def __init__(
        self,
        output_dir: Path,
        kind: str,
        compression: type[Compression],
        columns: Sequence[SchemaColumn],
        max_file_size_bytes: int = 128 * 1024 * 1024,
    ) -> None:
        super().__init__(output_dir, kind, compression, max_file_size_bytes)
        self.columns = columns


class NDJsonWriter(FileWriter[TextIOWrapper]):
    format = ".ndjson"

    class _DateTimeEncoder(json.JSONEncoder):
        def default(self, obj: object) -> object:
            if isinstance(obj, date | datetime):
                return obj.isoformat()
            return super().default(obj)

    def _create_writer(self, filepath: Path) -> TextIOWrapper:
        """Create a writer for the given file path."""
        return self.compression_cls(filepath).open("w")

    def _write(self, writer: TextIOWrapper, chunks: Iterable[Chunk]) -> None:
        writer.writelines(
            f"{json.dumps(chunk, cls=self._DateTimeEncoder)}{self.compression_cls.newline}" for chunk in chunks
        )


class YAMLBaseWriter(FileWriter[TextIOWrapper], ABC):
    def _create_writer(self, filepath: Path) -> TextIOWrapper:
        return self.compression_cls(filepath).open("w")

    def _write(self, writer: TextIOWrapper, chunks: Iterable[Chunk]) -> None:
        yaml.safe_dump_all(chunks, writer, sort_keys=False, explicit_start=True)


class YAMLWriter(YAMLBaseWriter):
    format = ".yaml"


class YMLWriter(YAMLBaseWriter):
    format = ".yml"


class CSVWriter(TableWriter[TextIOWrapper]):
    format = ".csv"

    def __init__(
        self,
        output_dir: Path,
        kind: str,
        compression: type[Compression],
        columns: Sequence[SchemaColumn],
        max_file_size_bytes: int = 128 * 1024 * 1024,
    ) -> None:
        super().__init__(output_dir, kind, compression, columns, max_file_size_bytes)
        self._csvwriter_by_file: dict[TextIOWrapper, csv.DictWriter] = {}

    def _create_writer(self, filepath: Path) -> TextIOWrapper:
        file = self.compression_cls(filepath).open("w")
        self._csvwriter_by_file[file] = self._create_dict_writer(file)
        return file

    def _write(self, writer: TextIOWrapper, chunks: Iterable[Chunk]) -> None:
        try:
            csv_writer = self._csvwriter_by_file[writer]
        except KeyError:
            csv_writer = self._create_dict_writer(writer)
            self._csvwriter_by_file[writer] = csv_writer
        csv_writer.writerows(self._prepare_row(row) for row in chunks)

    @staticmethod
    def _prepare_row(row: Chunk) -> dict[str, str | int | float | bool]:
        """Prepare a row for writing to CSV."""
        prepared_row = {}
        value: str | int | float | bool
        for col, cell in row.items():
            if isinstance(cell, list | dict):
                value = json.dumps(cell, cls=NDJsonWriter._DateTimeEncoder)
            elif isinstance(cell, date | datetime):
                value = cell.isoformat()
            elif cell is None:
                value = ""
            elif isinstance(cell, int | float | bool):
                value = cell
            else:
                value = str(cell)
            prepared_row[col] = value
        return prepared_row

    def _create_dict_writer(self, writer: IOBase) -> csv.DictWriter:
        csv_writer = csv.DictWriter(writer, fieldnames=[col.name for col in self.columns], extrasaction="ignore")
        csv_writer.writeheader()
        return csv_writer


class ParquetWriter(TableWriter["pq.ParquetWriter"]):
    format = ".parquet"

    def _create_writer(self, filepath: Path) -> "pq.ParquetWriter":
        import pyarrow.parquet as pq

        schema = self._create_schema()
        return pq.ParquetWriter(filepath, schema)

    def _write(self, writer: "pq.ParquetWriter", chunks: Iterable[Chunk]) -> None:
        import pyarrow as pa

        json_columns = self._json_columns()
        timestamp_columns = self._timestamp_columns()
        date_columns = self._date_columns()
        if not json_columns and not timestamp_columns and not date_columns:
            # If no special columns, we can write directly without processing
            table = pa.Table.from_pylist(chunks, schema=self._create_schema())
            writer.write_table(table)
            return

        processed_chunks: list[Chunk] = []
        for chunk in chunks:
            # Create a copy to avoid mutating the input, which is an unexpected side-effect.
            processed_chunk = chunk.copy()
            for col, cell_value in processed_chunk.items():
                if col in json_columns:
                    processed_chunk[col] = json.dumps(cell_value)
                elif col in timestamp_columns:
                    if isinstance(cell_value, list):
                        # MyPy fails to recognize that list of datetime and date are valid CellValues.
                        processed_chunk[col] = [self._to_datetime(value) for value in cell_value]  # type: ignore[assignment]
                    else:
                        processed_chunk[col] = self._to_datetime(cell_value)
                elif col in date_columns:
                    if isinstance(cell_value, list):
                        processed_chunk[col] = [self._to_date(value) for value in cell_value]  # type: ignore[assignment]
                    else:
                        processed_chunk[col] = self._to_date(cell_value)
            processed_chunks.append(processed_chunk)

        if not processed_chunks:
            return

        table = pa.Table.from_pylist(processed_chunks, schema=self._create_schema())
        writer.write_table(table)

    @lru_cache(maxsize=1)
    def _json_columns(self) -> set[str]:
        """Check if the writer supports JSON format."""
        return {col.name for col in self.columns if col.type == "json"}

    @lru_cache(maxsize=1)
    def _timestamp_columns(self) -> set[str]:
        """Check if the writer supports timestamp format."""
        return {col.name for col in self.columns if col.type == "timestamp"}

    @lru_cache(maxsize=1)
    def _date_columns(self) -> set[str]:
        return {col.name for col in self.columns if col.type == "date"}

    @classmethod
    def _to_datetime(cls, value: CellValue) -> CellValue:
        if isinstance(value, datetime) or value is None:
            output = value
        elif isinstance(value, date):
            output = datetime.combine(value, datetime.min.time())
        elif isinstance(value, int | float):
            # Assuming the value is a timestamp in milliseconds
            output = datetime.fromtimestamp(value / 1000.0)
        elif isinstance(value, str):
            output = cls._convert_data_modelling_timestamp(value)
        else:
            raise ToolkitTypeError(
                f"Unsupported value type for datetime conversion: {type(value)}. Expected datetime, date, int, float, or str."
            )
        if output is not None and output.tzinfo is None:
            # Ensure the datetime is in UTC
            output = output.replace(tzinfo=timezone.utc)
        elif output is not None and output.tzinfo is not None:
            # Convert to UTC if it has a timezone
            output = output.astimezone(timezone.utc)
        return output

    @classmethod
    def _to_date(cls, value: CellValue) -> CellValue:
        if isinstance(value, date) or value is None:
            return value
        elif isinstance(value, datetime):
            return value.date()
        elif isinstance(value, int | float):
            # Assuming the value is a timestamp in milliseconds
            return date.fromtimestamp(value / 1000.0)
        elif isinstance(value, str):
            return cls._convert_data_modelling_timestamp(value).date()
        else:
            raise ToolkitTypeError(
                f"Unsupported value type for date conversion: {type(value)}. Expected date, datetime, int, float, or str."
            )

    @classmethod
    def _convert_data_modelling_timestamp(cls, timestamp: str) -> datetime:
        """Convert a timestamp string from the data modeling format to a datetime object."""
        try:
            return datetime.fromisoformat(timestamp)
        except ValueError:
            # Typically hits if the timestamp has truncated milliseconds,
            # For example, "2021-01-01T00:00:00.17+00:00".
            # In Python 3.10, the strptime requires exact formats so we need both formats below.
            # In Python 3.11-13, if the timestamp matches on the second it will match on the first,
            # so when we set lower bound to 3.11 the loop will not be needed.
            for format_ in ["%Y-%m-%dT%H:%M:%S.%f%z", "%Y-%m-%dT%H:%M:%S%z"]:
                try:
                    return datetime.strptime(timestamp, format_)
                except ValueError:
                    continue
            raise ValueError(
                f"Invalid timestamp format: {timestamp}. Expected ISO 8601 format with optional milliseconds and timezone."
            )

    @lru_cache(maxsize=1)
    def _create_schema(self) -> "pa.Schema":
        """Create a pyarrow schema from the schema definition."""
        self._check_pyarrow_dependency()
        import pyarrow as pa

        fields: list[pa.Field] = []
        for prop in self.columns:
            pa_type = self._as_pa_type(prop.type, prop.is_array)
            fields.append(pa.field(prop.name, pa_type, nullable=True))
        return pa.schema(fields)

    @staticmethod
    def _check_pyarrow_dependency() -> None:
        if importlib.util.find_spec("pyarrow") is None:
            raise ToolkitMissingDependencyError(
                "Writing to parquet requires pyarrow. Install with 'pip install \"cognite-toolkit[table]\"'"
            )

    @staticmethod
    def _as_pa_type(type_: DataType, is_array: bool) -> "pa.DataType":
        """Convert a data type to a pyarrow type."""
        import pyarrow as pa

        if type_ == "string":
            pa_type = pa.string()
        elif type_ == "integer" or type_ == "epoch":
            pa_type = pa.int64()
        elif type_ == "float":
            pa_type = pa.float64()
        elif type_ == "boolean":
            pa_type = pa.bool_()
        elif type_ == "date":
            pa_type = pa.date32()
        elif type_ == "time":
            pa_type = pa.time64("ms")
        elif type_ == "json":
            pa_type = pa.string()
        elif type_ == "timestamp":
            pa_type = pa.timestamp("ms", tz="UTC")
        else:
            raise ToolkitValueError(f"Unsupported data type {type_}.")

        if is_array:
            pa_type = pa.list_(pa_type)
        return pa_type


FILE_WRITE_CLS_BY_FORMAT: Mapping[str, type[FileWriter]] = {}
TABLE_WRITE_CLS_BY_FORMAT: Mapping[str, type[TableWriter]] = {}
for subclass in get_concrete_subclasses(FileWriter):  # type: ignore[type-abstract]
    if not getattr(subclass, "format", None):
        continue
    if subclass.format in FILE_WRITE_CLS_BY_FORMAT:
        raise TypeError(
            f"Duplicate file format {subclass.format!r} found for classes "
            f"{FILE_WRITE_CLS_BY_FORMAT[subclass.format].__name__!r} and {subclass.__name__!r}."
        )
    # We know we have a dict, but we want to expose FILE_WRITE_CLS_BY_FORMAT as a Mapping
    FILE_WRITE_CLS_BY_FORMAT[subclass.format] = subclass  # type: ignore[index]
    if issubclass(subclass, TableWriter):
        if subclass.format in TABLE_WRITE_CLS_BY_FORMAT:
            raise TypeError(
                f"Duplicate table file format {subclass.format!r} found for classes "
                f"{TABLE_WRITE_CLS_BY_FORMAT[subclass.format].__name__!r} and {subclass.__name__!r}."
            )
        TABLE_WRITE_CLS_BY_FORMAT[subclass.format] = subclass  # type: ignore[index]



import sys
from types import MappingProxyType
from typing import Any, ClassVar, cast

from pydantic import Field, ModelWrapValidatorHandler, model_serializer, model_validator
from pydantic_core.core_schema import SerializationInfo, SerializerFunctionWrapHandler

from cognite_toolkit._cdf_tk.utils import humanize_collection

from .base import BaseModelResource, ToolkitResource

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


class Mapping(BaseModelResource):
    expression: str = Field(
        description="Custom transform expression written in the Cognite transformation language.", max_length=2000
    )


class MappingInput(BaseModelResource):
    type: ClassVar[str]

    @model_validator(mode="wrap")
    @classmethod
    def find_type(cls, data: "dict[str, Any] | MappingInput", handler: ModelWrapValidatorHandler[Self]) -> Self:
        if isinstance(data, MappingInput):
            return cast(Self, data)
        if not isinstance(data, dict):
            raise ValueError(f"Invalid input type data '{type(data)}' expected dict")

        if cls is not MappingInput:
            # We are already in a subclass, so just validate as usual
            return handler(data)
        # If not we need to find the right subclass based on the type field.
        if "type" not in data:
            raise ValueError("Invalid input data missing 'type' key")
        type_ = data["type"]
        if type_ not in _MAPPING_INPUT_CLS_BY_TYPE:
            raise ValueError(
                f"invalid type '{type_}'. Expected one of {humanize_collection(_MAPPING_INPUT_CLS_BY_TYPE.keys(), bind_word='or')}"
            )
        cls_ = _MAPPING_INPUT_CLS_BY_TYPE[type_]
        return cast(Self, cls_.model_validate({k: v for k, v in data.items() if k != "type"}))

    @model_serializer(mode="wrap", when_used="always", return_type=dict)
    def include_type(self, handler: SerializerFunctionWrapHandler) -> dict:
        serialized_data = handler(self)
        serialized_data["type"] = self.type
        return serialized_data


class JsonMappingInput(MappingInput):
    type = "json"


class XMLMappingInput(MappingInput):
    type = "xml"


class CSVMappingInput(MappingInput):
    type = "csv"
    delimiter: str = Field(
        description="A single ASCII character used as the separator in the CSV file.",
        min_length=1,
        max_length=1,
        default=",",
    )
    custom_keys: list[str] | None = Field(
        None,
        description="List of headers. If this is not set, the headers will be retrieved from the CSV file.",
        min_length=1,
        max_length=20,
    )


class ProtobufFile(BaseModelResource):
    file_name: str = Field(
        description="Name of protobuf file. Must contain only letters, numbers, underscores, and hyphens, and must end with '.proto'.",
        max_length=128,
        pattern=r"[a-zA-Z0-9_-]+\.proto",
    )
    content: str = Field(description="Protobuf file content. Must be a valid protobuf file.", max_length=10000)


class ProtobufMappingInput(MappingInput):
    type = "protobuf"
    message_name: str = Field(description="Name of root message in the protobuf files.", max_length=128)
    files: list[ProtobufFile] = Field(description="The protobuf schema in text format.", max_length=5000)


class HostedExtractorMappingYAML(ToolkitResource):
    external_id: str = Field(
        description="The external ID provided by the client. Must be unique for the resource type.", max_length=255
    )
    mapping: Mapping
    input: MappingInput | None = Field(None, description="The input format of the data to be transformed.")
    published: bool = Field(description="Whether this mapping is published and should be available to be used in jobs.")

    @model_serializer(mode="wrap")
    def serialize_input(self, handler: SerializerFunctionWrapHandler, info: SerializationInfo) -> dict:
        # Input is serialized as empty dict
        # This issue arises because Pydantic's serialization mechanism doesn't automatically
        # handle polymorphic serialization for subclasses of MappingInput.
        # To address this, we include the below to explicitly calling model dump on the input
        serialized_data = handler(self)
        if self.input is not None:
            serialized_data["input"] = self.input.model_dump(**vars(info))
        return serialized_data


_MAPPING_INPUT_CLS_BY_TYPE: MappingProxyType[str, type[MappingInput]] = MappingProxyType(
    {s.type: s for s in MappingInput.__subclasses__()}
)



from pathlib import Path

import pytest
from cognite.client.data_classes import (
    Transformation,
    TransformationDestination,
    TransformationNotification,
    TransformationNotificationWrite,
    TransformationSchedule,
    TransformationScheduleWrite,
    TransformationWrite,
)
from cognite.client.data_classes.transformations import NonceCredentials

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import DumpResourceCommand
from cognite_toolkit._cdf_tk.commands.dump_resource import TransformationFinder
from cognite_toolkit._cdf_tk.cruds import (
    TransformationCRUD,
    TransformationNotificationCRUD,
    TransformationScheduleCRUD,
)


@pytest.fixture(scope="session")
def deployed_transformation(toolkit_client: ToolkitClient) -> Transformation:
    transformation = TransformationWrite(
        external_id="toolkit_test_transformation",
        name="Toolkit Test Transformation",
        ignore_null_fields=True,
        query="SELECT * FROM _cdf.assets",
        destination=TransformationDestination.assets(),
        conflict_mode="abort",
        is_public=True,
    )
    existing = toolkit_client.transformations.retrieve(external_id=transformation.external_id)
    if existing:
        return existing
    destination_session = toolkit_client.iam.sessions.create()
    source_session = toolkit_client.iam.sessions.create()
    transformation.destination_nonce = NonceCredentials(
        destination_session.id, destination_session.nonce, toolkit_client.config.project
    )
    transformation.source_nonce = NonceCredentials(
        source_session.id, source_session.nonce, toolkit_client.config.project
    )
    return toolkit_client.transformations.create(transformation)


@pytest.fixture(scope="session")
def deployed_transformation_schedule(toolkit_client: ToolkitClient, deployed_transformation: Transformation) -> None:
    schedule = TransformationScheduleWrite(
        interval="0 12 * * 1",
        external_id=deployed_transformation.external_id,
        is_paused=True,
    )
    existing = toolkit_client.transformations.schedules.retrieve(external_id=schedule.external_id)
    if existing:
        return existing
    return toolkit_client.transformations.schedules.create(schedule)


@pytest.fixture(scope="session")
def deployed_transformation_notification(
    toolkit_client: ToolkitClient, deployed_transformation: Transformation
) -> None:
    notification = TransformationNotificationWrite(
        destination="my@example.com",
        transformation_external_id=deployed_transformation.external_id,
    )

    existing = toolkit_client.transformations.notifications.list(
        transformation_external_id=deployed_transformation.external_id, destination=notification.destination, limit=1
    )
    if existing:
        return existing[0]
    return toolkit_client.transformations.notifications.create(notification)


class TestDumpTransformation:
    def test_dump_transformation_with_schedule_and_notification(
        self,
        toolkit_client: ToolkitClient,
        deployed_transformation: Transformation,
        deployed_transformation_schedule: TransformationSchedule,
        deployed_transformation_notification: TransformationNotification,
        tmp_path: Path,
    ) -> None:
        cmd = DumpResourceCommand(silent=True)
        cmd.dump_to_yamls(
            TransformationFinder(toolkit_client, (deployed_transformation.external_id,)),
            output_dir=tmp_path,
            clean=False,
            verbose=False,
        )

        transformation_folder = tmp_path / "transformations"
        assert transformation_folder.exists()
        assert sum(1 for _ in transformation_folder.glob(f"*{TransformationCRUD.kind}.yaml")) == 1
        assert sum(1 for _ in transformation_folder.glob(f"*{TransformationScheduleCRUD.kind}.yaml")) == 1
        assert sum(1 for _ in transformation_folder.glob(f"*{TransformationNotificationCRUD.kind}.yaml")) == 1



from .cicd import get_cicd_environment
from .collection import flatten_dict, humanize_collection, in_dict, to_diff
from .file import (
    YAMLComment,
    YAMLWithComments,
    load_yaml_inject_variables,
    quote_int_value_by_key_in_yaml,
    read_yaml_content,
    read_yaml_file,
    safe_read,
    safe_write,
    sanitize_filename,
    stringify_value_by_key_in_yaml,
    tmp_build_directory,
)
from .graphql_parser import GraphQLParser
from .hashing import (
    calculate_directory_hash,
    calculate_hash,
    calculate_secure_hash,
)
from .modules import find_directory_with_subdirectories, iterate_modules, module_from_path, resource_folder_from_path
from .sentry_utils import sentry_exception_filter

__all__ = [
    "GraphQLParser",
    "YAMLComment",
    "YAMLWithComments",
    "calculate_directory_hash",
    "calculate_hash",
    "calculate_secure_hash",
    "find_directory_with_subdirectories",
    "flatten_dict",
    "get_cicd_environment",
    "humanize_collection",
    "in_dict",
    "iterate_modules",
    "load_yaml_inject_variables",
    "module_from_path",
    "quote_int_value_by_key_in_yaml",
    "read_yaml_content",
    "read_yaml_file",
    "resource_folder_from_path",
    "safe_read",
    "safe_read",
    "safe_write",
    "safe_write",
    "sanitize_filename",
    "sentry_exception_filter",
    "stringify_value_by_key_in_yaml",
    "tmp_build_directory",
    "to_diff",
]



from collections.abc import Hashable, Iterable, Sequence
from pathlib import Path
from typing import Any, final

from cognite.client.data_classes.capabilities import AppConfigAcl, Capability
from cognite.client.utils.useful_types import SequenceNotStr
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.search_config import (
    SearchConfig,
    SearchConfigList,
    SearchConfigWrite,
    SearchConfigWriteList,
    ViewId,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import SearchConfigYAML
from cognite_toolkit._cdf_tk.utils import sanitize_filename
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_identifiable, dm_identifier

from .datamodel import ViewCRUD


@final
class SearchConfigCRUD(ResourceCRUD[ViewId, SearchConfigWrite, SearchConfig, SearchConfigWriteList, SearchConfigList]):
    support_drop = False
    folder_name = "cdf_applications"
    filename_pattern = r"^.*SearchConfig$"
    resource_cls = SearchConfig
    resource_write_cls = SearchConfigWrite
    list_cls = SearchConfigList
    list_write_cls = SearchConfigWriteList
    yaml_cls = SearchConfigYAML
    dependencies = frozenset({ViewCRUD})
    kind = "SearchConfig"
    _doc_base_url = "https://api-docs.cogheim.net/redoc/#tag/"
    _doc_url = "Search-Config/operation/upsertSearchConfigViews"

    def __init__(self, client: ToolkitClient, build_path: Path | None, console: Console | None):
        super().__init__(client, build_path, console)
        self._existing_configs_cache: dict[ViewId, int] | None = None

    @property
    def display_name(self) -> str:
        return "search config"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[SearchConfigWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = [AppConfigAcl.Action.Read] if read_only else [AppConfigAcl.Action.Read, AppConfigAcl.Action.Write]

        return AppConfigAcl(
            actions=actions,
            scope=AppConfigAcl.Scope.AppConfig(apps=["SEARCH"]),
            allow_unknown=True,
        )

    @classmethod
    def get_id(cls, item: SearchConfig | SearchConfigWrite | dict) -> ViewId:
        if isinstance(item, dict):
            return ViewId.load(item.get("view", {}))
        return item.view

    @classmethod
    def dump_id(cls, id: ViewId) -> dict[str, Any]:
        return {"view": id.dump()}

    @classmethod
    def as_str(cls, id: ViewId) -> str:
        return sanitize_filename(f"{id.external_id}_{id.space}")

    def dump_resource(self, resource: SearchConfig, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        dumped.pop("id", None)
        for key in ["columnsLayout", "filterLayout", "propertiesLayout"]:
            if not dumped.get(key) and key not in local:
                dumped.pop(key, None)
        return dumped

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> SearchConfigWrite:
        loaded = SearchConfigWrite._load(resource)

        if self._existing_configs_cache is None:
            all_configs = self.client.search.configurations.list()
            self._existing_configs_cache = {config.view: config.id for config in all_configs if config.id is not None}

        if loaded.view in self._existing_configs_cache:
            loaded.id = self._existing_configs_cache[loaded.view]

        return loaded

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path in [("columnsLayout",), ("filterLayout",), ("propertiesLayout",)]:
            return diff_list_identifiable(local, cdf, get_identifier=lambda x: x.get("property"))
        elif json_path == ("view",):
            return diff_list_identifiable(local, cdf, get_identifier=dm_identifier)
        return super().diff_list(local, cdf, json_path)

    def _upsert(self, items: SearchConfigWrite | SearchConfigWriteList) -> SearchConfigList:
        """
        Upsert search configurations using the upsert method
        """
        if isinstance(items, SearchConfigWrite):
            items = SearchConfigWriteList([items])

        result = SearchConfigList([])
        for item in items:
            created = self.client.search.configurations.upsert(item)
            result.append(created)
        return result

    def create(self, items: SearchConfigWriteList) -> SearchConfigList:
        """
        Create new search configurations using the upsert method
        """
        return self._upsert(items)

    def retrieve(self, ids: SequenceNotStr[ViewId]) -> SearchConfigList:
        """Retrieve search configurations by their IDs"""
        all_configs = self.client.search.configurations.list()
        # The API does not support server-side filtering, so we filter in memory.
        return SearchConfigList([config for config in all_configs if config.view in ids])

    def update(self, items: SearchConfigWriteList) -> SearchConfigList:
        """
        Update search configurations using the upsert method
        """
        return self._upsert(items)

    def delete(self, ids: SequenceNotStr[ViewId]) -> int:
        """
        Delete is not implemented in the API client
        """
        raise NotImplementedError("Delete operation is not supported for search config")

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[SearchConfig]:
        """Iterate through all search configurations"""
        if space or data_set_external_id or parent_ids:
            # These filters are not supported for SearchConfig
            return iter([])

        all_configs = self.client.search.configurations.list()
        return iter(all_configs)



import threading
from dataclasses import dataclass, field


@dataclass
class ItemsRequestTracker:
    """Tracks the state of requests split from an original request.

    Attributes:
        max_failures_before_abort (int): Maximum number of allowed failed split requests before aborting
            the entire operation. A value of -1 indicates no early abort.
        lock (threading.Lock): A lock to ensure thread-safe updates to the failure count.
        failed_split_count (int): The current count of failed split requests.

    """

    max_failures_before_abort: int = -1  # -1 means no early abort
    lock: threading.Lock = field(default_factory=threading.Lock, init=False)
    failed_split_count: int = field(default=0, init=False)

    def register_failure(self) -> None:
        """Register a failed split request and return whether to continue splitting."""
        with self.lock:
            self.failed_split_count += 1

    def limit_reached(self) -> bool:
        """Check if the failure limit has been reached."""
        with self.lock:
            if self.max_failures_before_abort < 0:
                return False
            return self.failed_split_count >= self.max_failures_before_abort
